# bioreactor

## Important

- RPi.GPIO will not work on Raspberry Pi 5 which is used by several functions in the script. Alternatives will be developed for Raspberry Pi 5 for GPIO pin control

- Whenever pressure is over the critical pressure (const_critical_pressure), gas valve will be triggered and reactor operation will stop.

- When using Presens oxygen meter OXY1RS485, and the device comes with a pre-configured parity of EVEN, run parity_change() in scripts/oxygen_rs495.py to change parity to NONE before using.

## To-dos

## Hardware set-up

## Raspberry Pi configuration

- Follow this git protocol: [SOP_reactor_construction](https://git.ufz.de/kappel/SOP/-/blob/main/RaspberryPi_Reactor/SOP_reactor_construction/SOP_reactor_construction.md). If you do not have access to this, contact Chang Ding (chang.ding@ufz.de)

    - Activate one-wiring protocol, and connect respective GPIO pins of Raspberry Pi, if temperature sensor DS18B20 is used

    - Connect respective GPIO pins of Raspberry Pi to the 433MHz signal sender if it is to be used

## Dependency

- **Python-LJM and Labjack-M** (required if Labjack analog-to-digital converter, e.g., T7Pro, is used)

- **dropbox-uploader** (required if dropbox upload function is used)

- **raspberry-remote** (required if 433MHz signal sender is used to control on/off of device)

- **wiringpi** (required by raspberrypi-remote), which is preinstalled on a standard raspbian system. To update or install on a Raspbian-Lite system:

    ```
    sudo apt-get install wiringpi
    ```

- **psutil** #memory usage, CPU temperature, disk usage, etc

- **matplotlib** #see comment in class Plot, Raspberry Pi 4B is needed for the plotting

- **pandas**

- **serial** #pyserial

- **minimalmodbus** #for oxy1rs485

- **struct** #necessary for hex conversion if using Bronkhost flow meter

- **MCP342x** #necessary if using MCP3424 analog-digital converter

- **smbus** #necessary for i2c sensor (e.g., MCP3424, Sensirion flow sensor)

## Software Installation

- Pull the git project

    ```
    cd /home/pi/Desktop
    git clone https://git.ufz.de/dingc/bioreactor.git
    git submodule update --init
    ```
- If dropbox upload and download function is to be used

    - copy /home/pi/Desktop/bioreactor/software_required/Dropbox-Uploader to /home/pi/, and configure according to instruction to link to one dropbox account of your choice.

        ```
        mv /home/pi/Desktop/bioreactor/software_required/Dropbox-Uploader /home/pi/
        ```

- If the 433 MHz signal sender is to be used:

    - copy /home/pi/Desktop/bioreactor/raspberry-remote to /home/pi/, and compile using "make"
    
        ```
        mv /home/pi/Desktop/bioreactor/software_required/raspberry-remote /home/pi/
        cd /home/pi/raspberry-remote
        make
        ```
    - If compiling fails (i.e., no executable called "send" was generated), directly use the send executable in the directory of software_required
        - To do that: 

            ```
            cp /home/pi/Desktop/bioreactor/software_required/send /home/pi/raspberry-remote/
            chmod +x cd /home/pi/raspberry-remote/send
            ```
- If the Labjack analog/digital converter is to be used, e.g., T7Pro:

    - Compile and install Labjack-M

    - Install the python module for Lajback - Python_LJM_2020_11_20

- Copy the subfolder `config` and its content to `/home/pi/Desktop/`

- Check if any of the following folders/files are already available and should be used as the starting point. If yes, copy them to the corresponding path. Otherwise, they will be generated automatically as empty folders/files

    - '/home/pi/Desktop/screenshot', \ #store screenshot that will be uploaded to dropbox. Files older than 7 days will be deleted

    - '/home/pi/Desktop/log', \ #store reactor data

    - '/home/pi/Desktop/config', \ #store all important settings

    - '/home/pi/Desktop/plot', \ #store plots

    - '/home/pi/Desktop/plot/continuous', \

    - '/home/pi/Desktop/plot/once'

    - '/home/pi/Desktop/config/Program.txt', \ #store reactor programs

    - '/home/pi/Desktop/config/Slope.txt', \ #store pressure slope history

    - '/home/pi/Desktop/config/last_release.txt', \ #store the time of the latest gas release

    - '/home/pi/Desktop/config/teamviewer.txt', \ #communicate with dropbox, to reactivate teamviewer

    - '/home/pi/Desktop/config/upload.txt', \ #communicate with dropbox, to upload log data of current day to dropbox

    - '/home/pi/Desktop/config/warning_switch.txt', \ #communicate with dropbox, so that IFTTT sent an email reminder in case of emergency

    - '/home/pi/Desktop/config/warning_message.txt', \ #same above, the content of the warning message

    - '/home/pi/Desktop/config/Effluent.txt' \ #store the water level in the effluetn bottle, in mL

- Configure "/home/pi/Desktop/config/device_configuration_example.txt" according to the set-up and change the file name to "device_configuration.txt"

## Authors

- Chang Ding (main contributor), chang.ding@ufz.de

- Sophia Kunze (fixed memory leakage of the plotting module)