"""built-in modules"""
import time
import datetime
import re
import os
from subprocess import Popen, PIPE # these commands does not require shell=True: mkdir/cat/rm/dropbox_uploader/'sudo teamviewer daemon restart'
import glob #used in uploading files
import threading
from math import log #used in calculating doubling time
from tkinter import ttk
import tkinter as tk
import tkinter.messagebox as mbox
import serial
from serial.tools import list_ports

"""determine if raspberry pi is older than 4, relevant for plotting"""
def get_raspberry_pi_model():
    try:
        with open('/proc/device-tree/model', 'r') as f:
            model = f.read().strip()
            return model
    except FileNotFoundError:
        return None

model = get_raspberry_pi_model()
global flag_raspberrypi4_or_newer
flag_raspberrypi4_or_newer = True
if model:
    print("Raspberry Pi model: {}".format(model))
    if "Raspberry Pi 4" in model or "Raspberry Pi 5" in model:
        flag_raspberrypi4_or_newer = True
    else:
        flag_raspberrypi4_or_newer = False
else:
    print("Not running on a Raspberry Pi or unable to determine the model.")


#example for branching

"""
check required folders, files, software
The list has to be consistent with the README.md
"""
required_folder = ['/home/pi/Desktop/screenshot', \
                   '/home/pi/Desktop/log', \
                   '/home/pi/Desktop/config', \
                   '/home/pi/Desktop/plot', \
                   '/home/pi/Desktop/plot/continuous', \
                   '/home/pi/Desktop/plot/once']
for item in required_folder:
    if not os.path.exists(item):
        proc = Popen(['mkdir', item])
        proc.communicate()
required_file = ['/home/pi/Desktop/config/Program.txt', \
                 '/home/pi/Desktop/config/Slope.txt', \
                 '/home/pi/Desktop/config/last_release.txt', \
                 '/home/pi/Desktop/config/teamviewer.txt', \
                 '/home/pi/Desktop/config/upload.txt', \
                 '/home/pi/Desktop/config/warning_switch.txt', \
                 '/home/pi/Desktop/config/warning_message.txt', \
                 '/home/pi/Desktop/config/device_configuration.txt', \
                 '/home/pi/Desktop/config/Effluent.txt'] 
for item in required_file:
    if not os.path.exists(item):
        open(item, 'w').close()

"""Load and confirm device settings"""
global const_pump
global const_id_hamilton
global const_id_oxy1rs232
global const_id_oxy1rs485
global const_id_fibox
global const_meter_temperature
global const_meter_pressure
global const_meter_flow
global const_meter_DO1
global const_meter_DO2
global const_meter_DO1_comment
global const_meter_DO2_comment

""" AIN port on Labjack for pressure / redox """
global const_AIN_pressure_reactor
global const_AIN_pressure_medium
global const_AIN_redox

"""Sensor/controller addresses"""
global const_id_usb_hub
global const_DS18B20_address
global const_hamilton_port
global const_peristaltic_port
global const_fibox_port
global const_oxy1rs232_port
global const_oxy1rs485_port
global const_bronkhorst_port
global const_multiseven_port

"""address of Brennenstuhl RCS 1000N comfort remote switch"""
global const_valve_switch_1
global const_valve_switch_2
global const_heater_switch_1
global const_heater_switch_2
global const_stirring_switch_1
global const_stirring_switch_2
global const_stirring_interval

"""constant"""
global const_max_effluent_volume #size of effluent bottle (mL)
global const_critical_pressure #always release gas when reactor pressure is above this value (mV)
global const_syringe_size #hamilton syringe size in microliter
global const_port_nitrite
global const_port_ammonium
global const_port_sampling
global const_port_reactor_in
global const_port_reactor_out
global const_port_effluent
global const_default_program
global const_reactor_size

global const_nitrite_to_N2_ratio
global const_mV_per_mL_gas
global const_gas_temperature_for_calculation
global const_nitrite_in_mM_in_final_medium
global const_pressure_threshold
global const_pressure_release_in_seconds
global const_set_point_temperature

"""acceptable pH and temperature range, for filtering of raw values"""
global const_allowed_max_temperature
global const_allowed_min_temperature
global const_allowed_max_pH
global const_allowed_min_pH

with open('/home/pi/Desktop/config/device_configuration.txt','r') as f:
    config_text = f.read()
    lines = config_text.split("\n")
    for line in lines:
        item = line.split("\t")
        if item[0] == "pump":
            const_pump = item[1] # Hamilton or Peristaltic or -, used to decide whether to load reactor/control module
        elif item[0] == "id_hamilton":
            const_id_hamilton = item[1] #PID of the USB-RS232 converting cable of the Hamilton pump
        elif item[0] == "id_oxy1rs485":
            const_id_oxy1rs485 = item[1] #serial number of the oxygen meter RS485
        elif item[0] == "id_oxy1rs232":
            const_id_oxy1rs232 = item[1] #serial number of the oxygen meter RS232
        elif item[0] == "id_fibox":
            const_id_fibox = item[1] #PID of the USB-RS232 converting cable of the Fibox3Trace oxygen meter
        elif item[0] == "temperature_sensor":
            const_meter_temperature = item[1] # DS18B20 or Multiseven or -
        elif item[0] == "address_temperature_sensor":
            const_DS18B20_address = item[1] # address of the DS18B20 sensor, will be automatically updated
        elif item[0] == "pressure_sensor":
            const_meter_pressure = item[1] # Labjack or MCP3424 or -
        elif item[0] == "flow_meter":
            const_meter_flow = item[1] # Bronkhorst or -
        elif item[0] == "DO1_oxygen_meter":
            const_meter_DO1 = item[1] # Fibox (Fibox 3 Trace) or OXY1RS232 or OXY1RS485 or -
        elif item[0] == "DO1_comment":
            const_meter_DO1_comment = item[1] #position of sensor (liquid/headspace), type of sensor (PST6/PST9)
        elif item[0] == "DO2_oxygen_meter":
            const_meter_DO2 = item[1] # Fibox (Fibox 3 Trace) or OXY1RS232 or OXY1RS485 or -
        elif item[0] == "DO2_comment":
            const_meter_DO2_comment = item[1] #position of sensor (liquid/headspace), type of sensor (PST6/PST9)
        elif item[0] == "AIN_pressure_reactor":
            const_AIN_pressure_reactor = item[1] # AIN0/1/2 or -  
        elif item[0] == "AIN_pressure_medium":
            const_AIN_pressure_medium = item[1] # AIN0/1/2 or -  
        elif item[0] == "AIN_redox":
            const_AIN_redox = item[1] # currently can only be AIN0 or -, and AIN0 should be equipped with LJTick-InBuff 
        elif item[0] == "usb_port_hamilton":
            const_hamilton_port = item[1] # USB port in form of /dev/ttyUSB1 or -
        elif item[0] == "usb_port_peristaltic":
            const_peristaltic_port = item[1] # USB port in form of /dev/ttyUSB1 or -     
        elif item[0] == "usb_port_fibox":
            const_fibox_port = item[1] # USB port in form of /dev/ttyUSB1 or -
        elif item[0] == "usb_port_oxy1rs232":
            const_oxy1rs232_port = item[1] # USB port in form of /dev/ttyUSB1 or -
        elif item[0] == "usb_port_oxy1rs485":
            const_oxy1rs485_port = item[1] # USB port in form of /dev/ttyUSB1 or -
        elif item[0] == "usb_port_multiseven":
            const_multiseven_port = item[1] # USB port in form of /dev/ttyUSB1 or -
        elif item[0] == "usb_port_bronkhorst":
            const_bronkhorst_port = item[1] # USB port in form of /dev/ttyUSB1 or -
        elif item[0] == "id_usb_hub": # id of usb hub for resetting connection
            const_id_usb_hub = item[1]
        elif item[0] == "valve_switch_1":
            const_valve_switch_1 = item[1]
        elif item[0] == "valve_switch_2":
            const_valve_switch_2 = item[1]
        elif item[0] == "heater_switch_1":
            const_heater_switch_1 = item[1]
        elif item[0] == "heater_switch_2":
            const_heater_switch_2 = item[1]
        elif item[0] == "stirrer_switch_1":
            const_stirring_switch_1 = item[1]
        elif item[0] == "stirrer_switch_2":
            const_stirring_switch_2 = item[1]
        elif item[0] == "stirring_interval":
            const_stirring_interval = item[1]
        elif item[0] == "max_effluent_volume": 
            const_max_effluent_volume = int(item[1]) #size of effluent bottle (mL)
        elif item[0] == "critical_pressure":
            const_critical_pressure = int(item[1]) #always release gas when reactor pressure is above this value (mV)
        elif item[0] == "syringe_size":
            const_syringe_size = int(item[1]) #hamilton syringe size in microliter       
        elif item[0] == "syringe_port_nitrite":
            const_port_nitrite = item[1]
        elif item[0] == "syringe_port_ammonium":
            const_port_ammonium = item[1]
        elif item[0] == "syringe_port_sampling":
            const_port_sampling = item[1]
        elif item[0] == "syringe_port_reactor_in":
            const_port_reactor_in = item[1]
        elif item[0] == "syringe_port_reactor_out":
            const_port_reactor_out = item[1]
        elif item[0] == "syringe_port_effluent":
            const_port_effluent = item[1]
        elif item[0] == "allowed_max_temperature":
            const_allowed_max_temperature = float(item[1])
        elif item[0] == "allowed_min_temperature":
            const_allowed_min_temperature = float(item[1])
        elif item[0] == "allowed_max_pH":
            const_allowed_max_pH = float(item[1])
        elif item[0] == "allowed_min_pH":
            const_allowed_min_pH = float(item[1])
        elif item[0] == "default_program":
            const_default_program = item[1]
        elif item[0] == "reactor_size":
            const_reactor_size = int(item[1])
        elif item[0] == "nitrite_to_N2_ratio":
            const_nitrite_to_N2_ratio = float(item[1])
        elif item[0] == "mV_per_mL_gas":
            const_mV_per_mL_gas = float(item[1])
        elif item[0] == "gas_temperature_for_calculation":
            const_gas_temperature_for_calculation = float(item[1])
        elif item[0] == "nitrite_in_mM_in_final_medium":
            const_nitrite_in_mM_in_final_medium = float(item[1])
        elif item[0] == "pressure_threshold":
            const_pressure_threshold = int(item[1])
        elif item[0] == "pressure_release_in_seconds":
            const_pressure_release_in_seconds = int(item[1])
        elif item[0] == "set_point_temperature":
            const_set_point_temperature = float(item[1])

"""third party modules
Library versions that are confirmed to work on a Raspberry Pi 4B in 2025
psutil: 5.5.1
matplotlib: 3.5.2 this seems to be critical for certain functions
pandas: 1.3.5
minimalmodbus: 2.0.1
labjack-ljm: 1.21.0
"""
import psutil #memory usage, CPU temperature, disk usage, etc
if flag_raspberrypi4_or_newer: # Only import matplotlib if raspberry pi is 4 or newer, otherwise problem
    import matplotlib #see comment in class Plot, Raspberry Pi 4B is needed for the plotting
    matplotlib.use('TkAgg')
    import matplotlib.pyplot as plt
    #from matplotlib.figure import Figure
    from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg # NavigationToolbar2Tk #this is needed to embed the figure in tk GUI
from pandas import read_csv, concat, Timestamp #DataFrame
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters() #register the datetime converter to avoid warning
import minimalmodbus #for oxy1rs485
if const_meter_flow == 'Bronkhorst':
    import struct #necessary for hex conversion if using Bronkhost flow meter
if const_meter_pressure == 'Labjack':
    from labjack import ljm #necessary if using labjack analog-digital converter
elif const_meter_pressure == 'MCP3424':
    import MCP342x #necessary if using MCP3424 analog-digital converter
    import smbus #necessary for i2c sensor (e.g., MCP3424, Sensirion flow sensor)

if const_meter_temperature == 'DS18B20':
    import RPi.GPIO as GPIO #necessary if using DS18B20 temperature sensor
    GPIO.setmode(GPIO.BOARD)
    GPIO.setwarnings(False)
    global temperature_power_pin
    print("Looking for DS18B20 temperature sensor (takes ~10s)")
    temperature_power_pin = 15 #connect temperature sensor power (red) to GPIO15 (board mode)
    GPIO.setup(temperature_power_pin,GPIO.OUT)
    GPIO.output(temperature_power_pin, GPIO.HIGH)
    time.sleep(5)
    os.chdir('/sys/bus/w1/devices')
    sensor_address=''
    flag_search_temperature_sensor = True
    while flag_search_temperature_sensor == True:
        for child in glob.glob('28-*'): #sensor address starts with 28-
            root = tk.Tk() # so that there is no second window of tk
            root.overrideredirect(1)
            root.withdraw()
            result = mbox.askyesno('Temperature', 'Is the sensor address ' + child + '?')
            root.destroy()
            if result == True:
                with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
                    config_text = config_text.replace(("address_temperature_sensor\t" + const_DS18B20_address),("address_temperature_sensor\t" + child))
                    f.write(config_text)
                sensor_address = child
                flag_search_temperature_sensor = False
        if sensor_address == '':
            root = tk.Tk() # so that there is no second window of tk
            root.overrideredirect(1)
            root.withdraw()
            search = mbox.askyesno('Warning', 'No temperature sensor detected/confirmed. Perform the search again?')
            root.destroy()
            if search == True:
                print("Resetting power pin of temperature sensor (takes ~20s)")
                GPIO.output(temperature_power_pin, GPIO.LOW)
                time.sleep(20)
                GPIO.output(temperature_power_pin, GPIO.HIGH)
                pass
            else:
                flag_search_temperature_sensor = False
                const_meter_temperature = '-'

"""Other global parameters"""
global flag_release_gas_in_progress
flag_release_gas_in_progress = False

"""Below are automatic port determination"""
ports = list_ports.comports() #get all ports
"""determine Fibox port"""
port_fibox = '-'
for p in ports:
    if const_id_fibox in p.hwid:
        port_fibox = p.device
with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
    config_text = config_text.replace(("usb_port_fibox\t" + const_fibox_port),("usb_port_fibox\t" + port_fibox))
    f.write(config_text)
    const_fibox_port = port_fibox
"""determine OXY1RS232 port"""
port_oxy1rs232 = '-'
for p in ports:
    if const_id_oxy1rs232 in p.hwid:
        port_oxy1rs232 = p.device
with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
    config_text = config_text.replace(("usb_port_oxy1rs232\t" + const_oxy1rs232_port),("usb_port_oxy1rs232\t" + port_oxy1rs232))
    f.write(config_text)
    const_oxy1rs232_port = port_oxy1rs232
"""determine OXY1RS485 port"""
port_oxy1rs485 = '-'
for p in ports:
    if const_id_oxy1rs485 in p.hwid:
        port_oxy1rs485 = p.device
with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
    config_text = config_text.replace(("usb_port_oxy1rs485\t" + const_oxy1rs485_port),("usb_port_oxy1rs485\t" + port_oxy1rs485))
    f.write(config_text)
    const_oxy1rs485_port = port_oxy1rs485
"""determine Hamilton syringe port"""
port_hamilton = '-'
for p in ports:
    if const_id_hamilton in p.hwid:
        port_hamilton = p.device     
with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
    config_text = config_text.replace(("usb_port_hamilton\t" + const_hamilton_port),("usb_port_hamilton\t" + port_hamilton))
    f.write(config_text)
    const_hamilton_port= port_hamilton

   
def tail(f, lines):
    """read a file handler f (must be opened with b option """
    """for python 3, otherwise f.seek does not work from the file end), """
    """and output the last <lines> lines"""
    """originally by liuyix / GitHub https://gist.github.com/liuyix/dcb2f94abeeb3c258a75 """
    """modified for python 3"""
    total_lines_wanted = lines

    BLOCK_SIZE = 1024
    f.seek(0, 2)
    block_end_byte = f.tell()
    lines_to_go = total_lines_wanted
    block_number = -1
    blocks = [] # blocks of size BLOCK_SIZE, in reverse order startingException in thread Thread-4:
                # from the end of the file
    while lines_to_go > 0 and block_end_byte > 0:
        if (block_end_byte - BLOCK_SIZE > 0):
            # read the last block we haven't yet read
            f.seek(block_number*BLOCK_SIZE, 2)
            blocks.append(f.read(BLOCK_SIZE))
        else:
            # file too small, start from begining
            f.seek(0,0)
            # only read what was not read
            blocks.append(f.read(block_end_byte))
        lines_found = blocks[-1].count(b'\n') #convert \n to bytes
        lines_to_go -= lines_found
        block_end_byte -= BLOCK_SIZE
        block_number -= 1
    all_read_text = b''.join(reversed(blocks)).decode() #decode to convert to string
    return '\n'.join(all_read_text.splitlines()[-total_lines_wanted:])

def send_hamilton(command_text):
    """send command to the hamilton syringe pump"""
    global const_hamilton_port
    global const_id_hamilton
    ports = list_ports.comports() #get all ports
    port_hamilton = '-'
    for p in ports:
        if const_id_hamilton in p.hwid: 
            port_hamilton = p.device
    if port_hamilton == const_hamilton_port:
        pass
    else:
        with open('/home/pi/Desktop/config/device_configuration.txt','r') as f:
            config_text = f.read()
        with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
            config_text = config_text.replace(("usb_port_hamilton\t" + const_hamilton_port),("usb_port_hamilton\t" + port_hamilton))
            f.write(config_text)
            print(datetime.datetime.now().strftime('%Y/%m/%d %H:%M:%S') + ": Hamilton pump port has changed from " + const_hamilton_port + " to " + port_hamilton)
            const_hamilton_port= port_hamilton         
    ser = serial.Serial(const_hamilton_port,9600,timeout=1)
    command= '/1' + command_text + 'R\r'
    ser.write(str.encode(command))
    s = ser.read(100)
    ser.close
    try:
        s = s.decode('utf-8',errors='ignore')
    except AttributeError:
        pass
    return s

def is_ready():
    """return status of hamilton syringe, true when is ready"""
    try:
        s=send_hamilton('F')
        if s[3:4]=='0':
            return 1
        else:
            return 0
    except:
        print("send_hamilton('F') returned error")
        return 0
    
def wait_until_ready_update_syringe():
    """wait for hamilton syringe pump to be ready before doing next step
    ALSO update syringe position"""
    while is_ready()!=1:
        try:
            syringe_position()
        except Exception as error:
            print(repr(error))
        time.sleep(0.1)
def wait_until_ready():
    """wait for hamilton syringe pump to be ready before doing next step"""
    while is_ready()!=1:
        time.sleep(0.1)

#Scripts only for reactor running/sampling module
def move_syringe(max_speed, position):
    """Use this script ONLY within the reactor running / sampling script. Don't use when device is busy! 
    move the syringe at the specified maximum speed (5-5800), valve (1-6), and position (1-3000).
    Update current syringe position parameter
    it check readiness"""
    command='/1v150V{0}a{1}R\r'.format(int(max_speed), int(position))
    send_hamilton(command)
    wait_until_ready_update_syringe()
    if syringe_position() == position:
        return 1
    else:
        return 0

def move_valve(valve):
    """Use this script ONLY within the reactor running / sampling script. Don't use when device is busy!
    move to specified valve 1-6. Update current valve position parameter
    it checks readiness
    return 1 if valve position reached, otherwise 0"""
    send_hamilton('h2600{0}'.format(valve))
    value_current_valve.set(int(valve))
    wait_until_ready()
    if valve_position() == valve:
        return 1
    else:
        return 0

#Standalone version of the syringe/valve controlling scripts
def move_syringe_standalone(max_speed, position):
    """Perform check to see if device is busy. Can be used separately
    move the syringe at the specified maximum speed (5-5800), and position (1-3000).
    Update current syringe position parameter.
    it does not checks readiness (it is only used by the button 'move syringe' and it anyway cannot
    update the syringe position in GUI"""
    result = False
    if running_reactor == True:
        result = mbox.askyesno('Warning', 'Reactor is still running. Do you want to continue?')
    #perform action only when reactor not running or the reply for above is Yes
    if running_reactor == False or result == True:
        send_hamilton('/1v150V{0}a{1}R\r'.format(int(max_speed), int(position)))
        wait_until_ready_update_syringe()
        time_now = datetime.datetime.now()
        actual_position = syringe_position() #this updates the value_current_syringe automatically
        if actual_position == position:
            with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Move syringe to position (0-3000):\t' + str(position) + '\n')
                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Move syringe to position (0-3000):\t' + str(position))
        else:
            with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'tried to move syringe to position (0-3000):\t' + str(position) + ' but failed\n')
                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'current syringe position (0-3000):\t' + str(actual_position) + '\n')
                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'tried to move syringe to position (0-3000):\t' + str(position) + ' but failed' )
                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'current syringe position (0-3000):\t' + str(actual_position))
    return

def move_valve_standalone(position):
    """Perform check to see if device is busy. Can be used separately
    move to specified valve 1-6. Update current valve position parameter
    it checks readiness"""
    result = False
    if running_reactor == True:
        result = mbox.askyesno('Warning', 'Reactor is still running. Do you want to continue?')
    #perform action only when reactor not running or the reply for above is Yes
    if running_reactor == False or result == True:            
        send_hamilton('h2600{0}'.format(position))
        wait_until_ready()
        time_now = datetime.datetime.now()
        actual_position = valve_position()  #this updates the value_current_valve automatically
        if actual_position == position:
            with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Move valve to position:\t' + str(position) + '\n')
                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Move valve to position:\t' + str(position) )
        else:
            with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'tried to move valve to position:\t' + str(position) + ' but failed\n')
                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'current valve position:\t' + str(actual_position) + '\n')
                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'tried to move valve to position:\t' + str(position) + ' but failed' )
                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'current valve position:\t' + str(actual_position))
    return

def initialize_standalone():
    """initialize syringe to valve 6 and syringe position0. Also perform check to see if device is busy. Can be used separately"""
    result = False
    if running_reactor == True:
        result = mbox.askyesno('Warning', 'Reactor is still running. Do you want to continue?')
    #perform action only when reactor not running or the reply for above is Yes
    if running_reactor == False or result == True:
        send_hamilton('Z')
        wait_until_ready() #it actually checks overall readiness
        send_hamilton('h30001') #change to H mode so that enquiry for syringe and valve position is possible
        if valve_position() != 6 or syringe_position() != 0:
            mbox.showinfo('Error', 'Initialization error. Please check status.')
    return

def syringe_position():
    """query absolute syringe postion and updage on GUI. 0-3000 in PSD/4, at 3000 pluger is at lowest position"""
    s=send_hamilton('?')
    position = int(float(s[3:-3]))
    value_current_syringe.set(position)
    return position

def valve_position():
    """query valve postion in PSD/4 and update on GUI"""
    s=send_hamilton('?24000')
    position = int(float(s[3:4]))
    value_current_valve.set(position)
    return position

def pipet(valve_in, valve_out, volume, max_speed, check_syringe):
    """pipet volume in microliter from valve_in to valve_out, at the maximum speed specified"""
    if check_syringe == True:
        global const_syringe_size
        result = mbox.askyesno('Important', 'Do you confirm that the syringe size is ' + str(const_syringe_size))
        if result == False:
            return 1
        
    initial_position = syringe_position()
    
    if initial_position != 0:
        window=tk.Tk()
        window.wm_withdraw()
        mbox.showinfo('Error', 'Syringe position is not zero. Syringe at ' + str(initial_position))
        return 1

    if 1<=valve_in<=6 and 1<=valve_out<=6 and 5<=max_speed<=5800 and volume>0 and isinstance(volume,int):
        pass
    else:
        window=tk.Tk()
        window.wm_withdraw()
        mbox.showinfo('Error', 'Input parameters are out of boundary or volume is not integer.')    
        return 1

    print('Start pipetting ' + str(volume) + ' ul from port ' + str(valve_in) + ' to port ' \
        + str(valve_out) + ' with ' + str(const_syringe_size) + ' ul syringe at the speed of ' + str(max_speed))

    if volume<=const_syringe_size:
        move_valve(valve_in)
        
        move_syringe(max_speed,volume*3000/const_syringe_size)
        move_valve(valve_out)
        
        move_syringe(max_speed,0)
        print('Pipetting complete.')
    else:
        repeat = int(volume/const_syringe_size)
        pipetted = 0
        while repeat >0:
            move_valve(valve_in)           
            move_syringe(max_speed,3000)
            move_valve(valve_out)         
            move_syringe(max_speed,0)
            repeat -= 1
            pipetted += 1
            print(str(pipetted*const_syringe_size) + ' ul pipetted.')
        residue = volume % const_syringe_size
        if residue != 0:
            move_valve(valve_in)        
            move_syringe(max_speed,residue*3000/const_syringe_size)
            move_valve(valve_out)    
            move_syringe(max_speed,0)
            print('Pipetting complete')
        else:
            print('Pipetting complete.')
            
def start_reactor():
    def run_reactor():
        """get parameters"""
        global const_syringe_size
        global running_pressure
        result = mbox.askyesno('Important', 'Do you confirm that the syringe size is ' + str(const_syringe_size) + '?' \
                                + ' The maximum volume (uL) in the program must not exceed the syringe size!')
        if result == False:
            return 1
        else:
            result = mbox.askyesno('Important', 'Do you confirm that all needles are completely inserted into stoppers: \
                                including those of medium tanks and effluent bottle?')
            if result == False:
                return 1
            elif running_pressure==False:
                result = mbox.askyesno('Warning', 'Pressure monitoring is not activated. Do you still want to start?')
                if result == False:
                    return 1
            elif pressure_control.get() == False:
                result = mbox.askyesno('Warning', 'Pressure control is not activated. Do you still want to start?')
                if result == False:
                    return 1
            
        cycletime = value_current_cycletime.get() #either in seconds (e.g., 5184), or in DO:phase:seconds (e.g., DO:6360:1800)
        s=ttk.Style()
        s.configure('Red.TLabel', foreground='red')
        global run_signal #required
        run_signal = True
        global running_reactor # required
        global time_current_pressurev
        global const_max_effluent_volume
        global const_port_reactor_in
        global const_port_reactor_out
        if running_reactor == True:
            mbox.showinfo('Error', 'Reactor is still running.')
        else:
            initial_position = syringe_position()
            if initial_position != 0:
                mbox.showinfo('Error', 'Syringe position is not zero. Syringe at ' + str(initial_position))
                return 1
            running_reactor = True
            button_start_reactor.grid_remove()
            button_stop_reactor.grid()
            label_running.grid()            
            
            reactor_volume_change = 0 #calculate cumulative reactor_volume_change in the cycle, as buffer for value_reactorV_imbalance

            while run_signal and value_current_cycle.get() != 0:
                start = datetime.datetime.now()            
                """if labjack is running and the last data was older than 2 min, stop reactor"""
                time_current_pressure = time_current_pressurev.get()
                if time_current_pressure == '':
                    time_diff = 0
                else:
                    time_diff = datetime.datetime.now() - datetime.datetime.strptime(time_current_pressure, '%y/%m/%d %H:%M:%S')
                if running_pressure and time_diff.seconds > 120:
                    time_now = datetime.datetime.now()
                    with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                        f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Reactor stopped due to pressure sensor error\n')
                        print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Reactor stopped due to pressure sensor error')
                    break

                """if effluent bottle is full, stop reactor"""
                if effluent_volume.get() > const_max_effluent_volume:
                    time_now = datetime.datetime.now()
                    with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                        f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Reactor stopped due to full effluent bottle\n')
                        print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Reactor stopped due to full effluent bottle')
                    break

                """remaining cycle number -1, set end time and last cycle time"""
                if value_current_cycle.get() != -1:
                    value_current_cycle.set(value_current_cycle.get()-1)
                lastcycle.set('started at ' + start.strftime('%y/%m/%d %H:%M:%S'))
                
                if 'DO1' in cycletime:
                    try:
                        nextcycle.set('starts when DO1 phase is > ' + cycletime.split(':')[1])
                        end = start + datetime.timedelta(seconds=int(cycletime.split(':')[2])) #give time for DO to drop after reactor cycle (need to experiment and decide on this time delay)
                    except:
                        """stop reactor if error happens"""
                        time_now = datetime.datetime.now()
                        print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Error in cycletime input: ' + cycletime)
                        button_start_reactor.grid
                        label_running.grid_remove()
                        break
                elif 'DO2' in cycletime:
                    try:
                        nextcycle.set('starts when DO2 phase is > ' + cycletime.split(':')[1])
                        end = start + datetime.timedelta(seconds=int(cycletime.split(':')[2])) #give time for DO to drop after reactor cycle (need to experiment and decide on this time delay)
                    except:
                        """stop reactor if error happens"""
                        time_now = datetime.datetime.now()
                        print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Error in cycletime input: ' + cycletime)
                        button_start_reactor.grid
                        label_running.grid_remove()
                        break
                else:
                    try:
                        end = start + datetime.timedelta(seconds=int(cycletime))                 
                        nextcycle.set('starts at ' + end.strftime('%y/%m/%d %H:%M:%S'))
                    except:
                        """stop reactor if error happens"""
                        time_now = datetime.datetime.now()
                        print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Error in cycletime input: ' + cycletime)
                        button_start_reactor.grid
                        label_running.grid_remove()
                        break
                
                """reading and running the program"""
                i = 1
                flat_repeat_in_program = False
                line_repeat_start = 1 #actual line number of the line containing 'start', minimum is 1
                line_repeat_end = 1 #actual line number of the line containing 'end'
                total_repeat = 1 #total repeats decided by the line containing 'start'

                #get current DO2 value for the following step, so that DO values will not suddenly change within one cycle
                value_DO2_phase = 0 # initialize the value (phase cannot be 0)
                if time_current_DO2_phase.get() != '':
                        time_diff = datetime.datetime.now() - datetime.datetime.strptime(time_current_DO2_phase.get(), '%y/%m/%d %H:%M:%S')
                        if time_diff.seconds < 1800: #do not take the value if the value is too old
                            value_DO2_phase = value_current_DO2_phase.get()

                while i < 26:
                    p = eval('value_current_step'+str(i)).get()
                    if p == '': #discard all program steps after the first blank step
                        if flat_repeat_in_program: #if there is repeat in the program, set repeat back to original count
                            eval('value_current_step'+str(line_repeat_start)).set('start-'+str(total_repeat)+'-'+str(total_repeat))
                        break
                    else:
                        eval('current_entry_step'+str(i)).configure(foreground = 'red')
                    if 'start' in p: #start of the repeat, e.g., start-5-5, means run the lines between start and end for a total of 5 times
                        flat_repeat_in_program = True
                        line_repeat_start = i
                        total_repeat = int(p.split('-')[1])
                        if int(p.split('-')[2]) > 0: #if the remaining run is > 0
                            eval('value_current_step'+str(i)).set('start-'+str(total_repeat)+'-'+str(int(p.split('-')[2])-1))
                            eval('current_entry_step'+str(i)).configure(foreground = 'black')
                        else: #no remaining run, jump to after line_repeat_end
                            eval('current_entry_step'+str(i)).configure(foreground = 'black')
                            i = line_repeat_end
                    elif 'end' in p: #end of the repeat
                        line_repeat_end = i
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                        i = line_repeat_start - 1 #jump to line_repeat_start
                    elif 'D<' in p: # trigger this line only when DO2 phase is lower than a threshold, e.g., D<:3000:200:2:500:4:0
                        q = p.split(':')
                        if float(value_DO2_phase) != 0 and float(value_DO2_phase) < int(q[1]): # if lower, then perform the normal cycle
                            while flag_releasing_gas.get() == True:
                                time.sleep(3) #wait until gas release is done
                                time_now = datetime.datetime.now()
                                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Waiting for gas release to finish')
                            """calculate flow direction for reactor volume change"""
                            flow_direction = 0
                            if int(q[3]) == const_port_reactor_in or int(q[3]) == const_port_reactor_out:
                                flow_direction = flow_direction + 1
                            if int(q[5]) == const_port_reactor_in or int(q[5]) == const_port_reactor_out:
                                flow_direction = flow_direction - 1
                            """calculate reactor volume change"""
                            reactor_volume_change = value_reactorV_imbalance.get()
                            reactor_volume_change = int(reactor_volume_change) + (int(q[6]) - int(q[4])) * flow_direction #uL
                            value_reactorV_imbalance.set(-100) #set to a non-zero value to prevent gas valve from opening
                            """perform pumping"""
                            time_now = datetime.datetime.now()
                            move_valve(int(q[3]))
                            move_syringe(q[2], int(int(q[4])*3000/const_syringe_size))
                            move_valve(int(q[5]))
                            move_syringe(q[2], int(int(q[6])*3000/const_syringe_size))
                            value_reactorV_imbalance.set(reactor_volume_change) #update value_reactorV_imbalance
                            with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:
                                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + p + '\n')
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                    elif 'D>' in p: # trigger this line only when DO2 phase is higher than a threshold, e.g., D<:5000:200:5:5000:1:0
                        q = p.split(':')
                        if float(value_DO2_phase) != 0 and float(value_DO2_phase) > int(q[1]): # if lower, then perform the normal cycle
                            while flag_releasing_gas.get() == True:
                                time.sleep(3) #wait until gas release is done
                                time_now = datetime.datetime.now()
                                print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Waiting for gas release to finish')
                            """calculate flow direction for reactor volume change"""
                            flow_direction = 0
                            if int(q[3]) == const_port_reactor_in or int(q[3]) == const_port_reactor_out:
                                flow_direction = flow_direction + 1
                            if int(q[5]) == const_port_reactor_in or int(q[5]) == const_port_reactor_out:
                                flow_direction = flow_direction - 1
                            """calculate reactor volume change"""
                            reactor_volume_change = value_reactorV_imbalance.get()
                            reactor_volume_change = int(reactor_volume_change) + (int(q[6]) - int(q[4])) * flow_direction #uL
                            value_reactorV_imbalance.set(-100) #set to a non-zero value to prevent gas valve from opening
                            """perform pumping"""
                            time_now = datetime.datetime.now()
                            move_valve(int(q[3]))
                            move_syringe(q[2], int(int(q[4])*3000/const_syringe_size))
                            move_valve(int(q[5]))
                            move_syringe(q[2], int(int(q[6])*3000/const_syringe_size))
                            value_reactorV_imbalance.set(reactor_volume_change) #update value_reactorV_imbalance
                            with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:
                                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + p + '\n')
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                    elif 'balance' in p: #force reactor volume balance for special cases
                        value_reactorV_imbalance.set(0)
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                    elif 'w' in p: #e.g., w10, wait for 10s
                        time.sleep(int(p[1:]))
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                    elif 'g' in p: #g for gas release, e.g., g15, release gas for 15 seconds
                        global flag_release_gas_in_progress
                        if flag_release_gas_in_progress == False:
                            release_gas(int(p[1:])+5, False) #give 5 more seconds for labjack to update the initial pressure
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                    elif 's' in p: #s for slow, e.g., s5:2:5000:4:0, withdraw at 500 speed (not shown), dispense at 5 speed
                        p = p[1:] #remove 's'
                        q = p.split(':')
                        while flag_releasing_gas.get() == True:
                            time.sleep(3) #wait until gas release is done
                            time_now = datetime.datetime.now()
                            print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Waiting for gas release to finish')
                        """calculate flow direction for reactor volume change"""
                        flow_direction = 0
                        if int(q[1]) == const_port_reactor_in or int(q[1]) == const_port_reactor_out:
                            flow_direction = flow_direction + 1
                        if int(q[3]) == const_port_reactor_in or int(q[3]) == const_port_reactor_out:
                            flow_direction = flow_direction - 1
                        """calculate reactor volume change"""
                        reactor_volume_change = value_reactorV_imbalance.get()
                        reactor_volume_change = int(reactor_volume_change) + (int(q[4]) - int(q[2])) * flow_direction #uL
                        value_reactorV_imbalance.set(-100) #set to a non-zero value to prevent gas valve from opening
                        """perform pumping"""
                        time_now = datetime.datetime.now()
                        move_valve(int(q[1]))                 
                        move_syringe(500, int(int(q[2])*3000/const_syringe_size)) #withdraw speed is always 500 (not p)                   
                        move_valve(int(q[3]))                  
                        move_syringe(q[0], int(int(q[4])*3000/const_syringe_size))                        
                        value_reactorV_imbalance.set(reactor_volume_change) #update value_reactorV_imbalance
                        with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                            f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + p + '\n')
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                    else: #normal program, with the first number being the syringe speed
                        q = p.split(':')
                        while flag_releasing_gas.get() == True:
                            time.sleep(3) #wait until gas release is done
                            time_now = datetime.datetime.now()
                            print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Waiting for gas release to finish')
                        """calculate flow direction for reactor volume change"""
                        flow_direction = 0
                        if int(q[1]) == const_port_reactor_in or int(q[1]) == const_port_reactor_out:
                            flow_direction = flow_direction + 1
                        if int(q[3]) == const_port_reactor_in or int(q[3]) == const_port_reactor_out:
                            flow_direction = flow_direction - 1
                        """calculate reactor volume change"""
                        reactor_volume_change = value_reactorV_imbalance.get()
                        reactor_volume_change = int(reactor_volume_change) + (int(q[4]) - int(q[2])) * flow_direction #uL
                        value_reactorV_imbalance.set(-100) #set to a non-zero value to prevent gas valve from opening
                        """perform pumping"""
                        time_now = datetime.datetime.now()
                        move_valve(int(q[1]))
                        move_syringe(q[0], int(int(q[2])*3000/const_syringe_size))
                        move_valve(int(q[3]))
                        move_syringe(q[0], int(int(q[4])*3000/const_syringe_size))
                        value_reactorV_imbalance.set(reactor_volume_change) #update value_reactorV_imbalance
                        with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:
                            f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + p + '\n')
                        eval('current_entry_step'+str(i)).configure(foreground = 'black')
                    i = i + 1
                """update effluent volume"""
                effluent = float(effluent_volume.get()) + input_flow.get()/1000
                effluent_volume.set(str(effluent))
                """monitor and wait"""
                flag = 1
                while flag == 1:
                    cycletime = value_current_cycletime.get()
                    cycle = value_current_cycle.get()
                    if run_signal == False or cycle == 0:
                        flag = 0
                    elif 'DO1' in cycletime:
                        nextcycle.set('starts when DO1 phase is > ' + cycletime.split(':')[1])
                        end = start + datetime.timedelta(seconds=int(cycletime.split(':')[2])) #give time for DO to drop after reactor cycle (need to experiment and decide on this time delay)
                        if time_current_DO1_phase.get() != '':
                            time_diff = datetime.datetime.now() - datetime.datetime.strptime(time_current_DO1_phase.get(), '%y/%m/%d %H:%M:%S')
                            if time_diff.seconds < 1800: #do not jump out of the loop if the value is too old
                                if datetime.datetime.now() > end and float(value_current_DO1_phase.get()) > int(cycletime.split(":")[1]): #only jump out when phase is higher than set value.
                                    flag = 0
                    elif 'DO2' in cycletime:
                        nextcycle.set('starts when DO2 phase is > ' + cycletime.split(':')[1])
                        end = start + datetime.timedelta(seconds=int(cycletime.split(':')[2])) #give time for DO to drop after reactor cycle (need to experiment and decide on this time delay)
                        if time_current_DO2_phase.get() != '':
                            time_diff = datetime.datetime.now() - datetime.datetime.strptime(time_current_DO2_phase.get(), '%y/%m/%d %H:%M:%S')
                            if time_diff.seconds < 1800: #do not jump out of the loop if the value is too old
                                if datetime.datetime.now() > end and float(value_current_DO2_phase.get()) > int(cycletime.split(":")[1]): #only jump out when phase is higher than set value.
                                    flag = 0
                    else:
                        end = start + datetime.timedelta(seconds=int(cycletime))               
                        nextcycle.set('starts at ' + end.strftime('%y/%m/%d %H:%M:%S'))
                        if datetime.datetime.now() > end:
                            flag = 0
                    time.sleep(1)
            running_reactor = False
            time_now = datetime.datetime.now()
            with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'reactor was stopped.\n')
            print(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'reactor was stopped.')
            label_running.grid_remove()
            label_stopping.grid_remove()
            button_start_reactor.grid()
    """transfer the input to current"""
    value_current_cycletime.set(input_cycletime.get())
    value_current_cycle.set(int(input_cycle.get()))
    for i in range(0,22):
        eval('value_current_step'+str(i+1)).set(eval('input_step'+str(i+1)).get())
    thread_reactor = threading.Thread(target=run_reactor) #, args=(cycletime, cycle, program)
    thread_reactor.start()
       
def stop_reactor():
    global run_signal #required
    run_signal = False
    button_stop_reactor.grid_remove()
    label_stopping.grid()
    label_running.grid_remove()

def load_program():
    flag = False
    program_selected = input_program.get()
    with open('/home/pi/Desktop/config/Program.txt','r') as f:
        for line in f:
            words = line.split('\t')
            if words[0].strip() == program_selected and flag == False:
                flag = True
            elif '#' in words[0]:
                flag = False
            elif flag == True:
                if words[0]=='cycletime':
                    input_cycletime.set(words[1].strip())
                if words[0]=='cycle':
                    input_cycle.set(words[1].strip())
                if words[0]=='step1':
                    input_step1.set(words[1].strip())
                if words[0]=='step2':
                    input_step2.set(words[1].strip())
                if words[0]=='step3':
                    input_step3.set(words[1].strip())
                if words[0]=='step4':
                    input_step4.set(words[1].strip())
                if words[0]=='step5':
                    input_step5.set(words[1].strip())
                if words[0]=='step6':
                    input_step6.set(words[1].strip())
                if words[0]=='step7':
                    input_step7.set(words[1].strip())
                if words[0]=='step8':
                    input_step8.set(words[1].strip())
                if words[0]=='step9':
                    input_step9.set(words[1].strip())
                if words[0]=='step10':
                    input_step10.set(words[1].strip())
                if words[0]=='step11':
                    input_step11.set(words[1].strip())
                if words[0]=='step12':
                    input_step12.set(words[1].strip())
                if words[0]=='step13':
                    input_step13.set(words[1].strip())
                if words[0]=='step14':
                    input_step14.set(words[1].strip())
                if words[0]=='step15':
                    input_step15.set(words[1].strip())
                if words[0]=='step16':
                    input_step16.set(words[1].strip())
                if words[0]=='step17':
                    input_step17.set(words[1].strip())
                if words[0]=='step18':
                    input_step18.set(words[1].strip())
                if words[0]=='step19':
                    input_step19.set(words[1].strip())
                if words[0]=='step20':
                    input_step20.set(words[1].strip())
                if words[0]=='step21':
                    input_step21.set(words[1].strip())
                if words[0]=='step22':
                    input_step22.set(words[1].strip())
                if words[0]=='syringe':
                    global const_syringe_size
                    if const_syringe_size != int(words[1].strip()):
                        mbox.askyesno('Warning', 'The actual syringe size is ' + str(const_syringe_size) + ' microliter, not ' + words[1].strip() + '!')
                if words[0]=='flow':
                    input_flow.set(words[1].strip())

def send_fibox(command_text):
    """send command to Presens Fibox3 Trace v3 oxygen meter"""
    global const_fibox_port
    ser = serial.Serial(port=const_fibox_port,baudrate=19200, bytesize=8, parity='N', stopbits=1, timeout=1)
    command= command_text + chr(13)
    ser.write(str.encode(command))
    s = ''
    count = 0
    while s == '' and count < 5:
        s = ser.read(100)
        try:
            s = s.decode('utf-8',errors='ignore')
        except AttributeError:
            pass
        time.sleep(0.1)
        count = count + 1
    ser.close
    return s

def read_fibox():
    """let oxy1rs232 read and return a string that contains amplitude, phase, and ppb"""
    #dict_DO_ppb = {} #read calibration into dictionary
    #with open('/home/pi/Desktop/config/Calibration_Sensor2008.txt','r') as f:
    #    for line in f:
    #       (raw, output) = line.split()
    #       dict_DO_ppb[float(raw)] = output
    raw_output = send_fibox('.').split(';')
    amplitude = int(raw_output[0][1:])
    phase = int(raw_output[1][1:])
    ppb = 0 #float(dict_DO_ppb[phase/100]), 0 as dummy number when use sensor other than PST6_2008, -1 causes error
    return str(amplitude) + "-" + str(phase) + "-" + str(ppb)
                
def send_oxy1rs232(command_text):
    """send command to Presens OXY-1 (RS232 version) oxygen meter"""
    global const_oxy1rs232_port
    global const_id_oxy1rs232
    ports = list_ports.comports() #get all ports
    port_oxy1rs232 = '-'
    for p in ports:
        if const_id_oxy1rs232 in p.hwid:
            port_oxy1rs232 = p.device
    if port_oxy1rs232 == const_oxy1rs232_port:
        pass
    else:
        with open('/home/pi/Desktop/config/device_configuration.txt','r') as f:
            config_text = f.read()
        with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
            config_text = config_text.replace(("usb_port_oxyrs232\t" + const_oxy1rs232_port),("usb_port_oxyrs232\t" + port_oxy1rs232))
            f.write(config_text)
            print(datetime.datetime.now().strftime('%Y/%m/%d %H:%M:%S') + ": OXY1RS232 port has changed from " + const_oxy1rs232_port + " to " + port_oxy1rs232)
            const_oxy1rs232_port= port_oxy1rs232
    ser = serial.Serial(port=const_oxy1rs232_port,baudrate=19200, bytesize=8, parity='N', stopbits=1, timeout=1)
    command= command_text + chr(13)
    ser.write(str.encode(command))
    s = ''
    count = 0
    time.sleep(0.2)
    while s == '' and count < 5:
        s = ser.read(2000)
        try:
            s = s.decode('utf-8',errors='ignore')
        except AttributeError:
            pass
        time.sleep(0.2)
        count = count + 1
    ser.close
    return s

def read_oxy1rs232():
    """let oxy1rs232 read and return a string that contains amplitude, phase, and ppb"""
    raw_output = send_oxy1rs232('data').split(';')
    amplitude = int(raw_output[1][1:])
    phase = int(raw_output[2][1:])
    ppb = round(int(raw_output[4][1:])/10,3) #PST9 (sens0006) only outputs ppmv (malt0001 for dry, oxyu0006 for ppmg, so "int(raw_output[4][1:])/10000 / 26.58" to obtain ug/L at 30 degree C
    #for pst6 (sens0003) and ppb, only "int(raw_output[4][1:])/10" (to change to ppb: malt0000 for humid, oxyu0007 for ppb)
    return str(amplitude) + "-" + str(phase) + "-" + str(ppb)

def read_oxy1rs485():
    """let oxy1rs485 read and return a string that contains amplitude, phase, and ppb"""
    global const_oxy1rs485_port
    global const_id_oxy1rs485
    ports = list_ports.comports() #get all ports
    port_oxy1rs485 = '-'
    for p in ports:
        if const_id_oxy1rs485 in p.hwid:
            port_oxy1rs485 = p.device
    if port_oxy1rs485 == const_oxy1rs485_port:
        pass
    else:
        with open('/home/pi/Desktop/config/device_configuration.txt','r') as f:
            config_text = f.read()
        with open('/home/pi/Desktop/config/device_configuration.txt','w') as f:
            config_text = config_text.replace(("usb_port_oxyrs485\t" + const_oxy1rs485_port),("usb_port_oxyrs485\t" + port_oxy1rs485))
            f.write(config_text)
            print(datetime.datetime.now().strftime('%Y/%m/%d %H:%M:%S') + ": OXY1RS485 port has changed from " + const_oxy1rs485_port + " to " + port_oxy1rs485)
            const_oxy1rs485_port= port_oxy1rs485    
    instrument = minimalmodbus.Instrument(const_oxy1rs485_port,1) #1 is Modbus ID
    instrument.serial.parity = "N"
    instrument.serial.timeout = 0.2
    instrument.write_registers(9000, [0x04,0x00]) #Trigger single measurement, see manual EOM-_t_O2-FOM_Modbus_RTU_Communication_Protocol
    time.sleep(1) #1s waiting is necessary for measurement to finish
    amplitude = instrument.read_float(4899,byteorder = 3)
    phase = instrument.read_float(4901,byteorder = 3)*100
    ppb = instrument.read_float(4905,byteorder = 3)
    instrument.serial.close() # the port needs to be closed. Otherwise "error from flowcontrol urb" occurs (dmesg | grep tty) after some time, and usb port changes
    return str(round(amplitude,0)) + "-" + str(round(phase,3)) + "-" + str(round(ppb,3))

def start_do1():
    """start the thread to read from meter that is assigned to DO1, update the current oxygen amplitude/phase and
    write to Oxygen1_log"""
    def read_do1():
        button_start_do1.grid_remove()
        button_stop_do1.grid()
        global running_do1
        running_do1 = True    
        while running_do1:
            time_start = datetime.datetime.now()
            string_time_start = time_start.strftime('%Y/%m/%d %H:%M:%S')
            string_time_start_short = time_start.strftime('%y/%m/%d %H:%M:%S')
            try:
                if const_meter_DO1 == "Fibox":
                    result = read_fibox().split('-')
                elif const_meter_DO1 == "OXY1RS232":
                    result = read_oxy1rs232().split('-')
                elif const_meter_DO1 == "OXY1RS485":
                    result = read_oxy1rs485().split('-')
                amplitude = float(result[0])
                phase = float(result[1])
                ppb = float(result[2]) 
                value_current_DO1_amplitude.set(amplitude)
                value_current_DO1_phase.set(phase)
                value_current_DO1_ppb.set(ppb)
                time_current_DO1_amplitude.set(string_time_start_short)
                time_current_DO1_phase.set(string_time_start_short)
                time_current_DO1_ppb.set(string_time_start_short)
                do1_comment = input_do1_comment.get()
                with open('/home/pi/Desktop/Oxygen1_log_' + time_start.strftime('%Y-%m-%d') + '.txt','a+') as f:
                    f.write('%s\tamplitude\t%d\tphase\t%d\tppb\t%f\t%s\n' % (string_time_start,amplitude,phase,ppb,do1_comment))                                    
            except Exception as error:
                print(string_time_start + '\tDO1 meter error\n' + str(repr(error)))
                global const_id_usb_hub
                if const_id_usb_hub != "-": # reset if the id is not empty (-)
                    print("reset usb hub connection")
                    Popen(['sudo', 'usbreset', const_id_usb_hub], stdout=PIPE)            
            do1_delay = 0 #intialize the value
            try:
                do1_delay = input_do1_delay.get()
            except:
                do1_delay = 100
            time_now = datetime.datetime.now()
            time_end = time_now + datetime.timedelta(seconds=do1_delay)
            while datetime.datetime.now() < time_end:
                time.sleep(3)
                if running_do1:
                    try:
                        do1_delay = input_do1_delay.get()
                    except:
                        do1_delay = do1_delay + 4 #make it longer than the 3s sleep
                else:
                    do1_delay = 0
                time_end = time_now + datetime.timedelta(seconds=do1_delay)
        button_start_do1.grid()
        button_stop_do1.grid_remove()
    thread_do1 = threading.Thread(target=read_do1)
    thread_do1.start()

def stop_do1():
    """stop the thread created by start_do1"""
    global running_do1
    running_do1 = False

def start_do2():
    """start the thread to read from oxygen meter that is assigned to DO2, update the current oxygen amplitude/phase and
    write to Oxygen2_log"""
    """use send_oxy1rs232('repo') to check sensor type. if need to change, sens0003 for pst6, sens0006 for pst9
    if pst9 is used, keep the /26.58 part of the ppb calculation, otherwise remove that
    malt0000 for humid measurement (ppb/L output in combination with oxyu0007)
    malt0001 for dry measurement (ppmg g output in combination with oxyu0006)"""
    def read_do2():
        button_start_do2.grid_remove()
        button_stop_do2.grid()
        global const_meter_DO2
        global running_do2
        running_do2 = True
        while running_do2:
            time_start = datetime.datetime.now()
            string_time_start = time_start.strftime('%Y/%m/%d %H:%M:%S')
            string_time_start_short = time_start.strftime('%y/%m/%d %H:%M:%S')
            try:
                if const_meter_DO2 == "Fibox":
                    result = read_fibox().split('-')
                elif const_meter_DO2 == "OXY1RS232":
                    result = read_oxy1rs232().split('-')
                elif const_meter_DO2 == "OXY1RS485":
                    result = read_oxy1rs485().split('-')
                amplitude = float(result[0])
                phase = float(result[1])
                ppb = float(result[2]) 
                value_current_DO2_amplitude.set(amplitude)
                value_current_DO2_phase.set(phase)
                value_current_DO2_ppb.set(ppb)
                time_current_DO2_amplitude.set(string_time_start_short)
                time_current_DO2_phase.set(string_time_start_short)
                time_current_DO2_ppb.set(string_time_start_short)
                do2_comment = input_do2_comment.get()
                with open('/home/pi/Desktop/Oxygen2_log_' + time_start.strftime('%Y-%m-%d') + '.txt','a+') as f:
                    f.write('%s\tamplitude\t%d\tphase\t%d\tppb\t%f\t%s\n' % (string_time_start, amplitude, phase, ppb, do2_comment))                                    
            except Exception as error:
                print(string_time_start + '\tDO2 meter error\n' + str(repr(error)))
                global const_id_usb_hub
                if const_id_usb_hub != "-": # reset if the id is not empty (-)
                    print("reset usb hub connection")
                    Popen(['sudo', 'usbreset', const_id_usb_hub], stdout=PIPE)
            do2_delay = 0 #intialize the value
            try:
                do2_delay = input_do2_delay.get()
            except:
                do2_delay = 100
            time_now = datetime.datetime.now()
            time_end = time_now + datetime.timedelta(seconds=do2_delay)
            while datetime.datetime.now() < time_end:
                time.sleep(3)
                if running_do2:
                    try:
                        do2_delay = input_do2_delay.get()
                    except:
                        do2_delay = do2_delay + 4 #make it longer than the 3s sleep
                else:
                    do2_delay = 0
                time_end = time_now + datetime.timedelta(seconds=do2_delay)
        button_start_do2.grid()
        button_stop_do2.grid_remove()
    thread_do2 = threading.Thread(target=read_do2)
    thread_do2.start()
def stop_do2():
    """stop the thread created by start_do2"""
    global running_do2
    running_do2 = False

"""temperature thread"""
def send_multiseven():
    """read buffer from multiseven, 100 byte"""
    global const_multiseven_port
    ser = serial.Serial(const_multiseven_port,9600,timeout=1)
    s=ser.read(100)
    ser.close
    try:
        s = s.decode('utf-8',errors='ignore')
    except AttributeError:
        pass
    return s

def start_temperature():
    """start the thread to read from temperature sensor, update the current T and
    write to Temperature_log"""
    def read_temperature():
        button_start_temperature.grid_remove()
        button_stop_temperature.grid()
        global running_temperature
        running_temperature = True
        global const_heater_switch_1
        global const_heater_switch_2
        global const_meter_temperature
        global const_allowed_max_temperature
        global const_allowed_min_temperature
        global const_allowed_max_pH
        global const_allowed_min_pH
        if const_meter_temperature == 'DS18B20':
            global temperature_power_pin
            global const_DS18B20_address        
        while running_temperature:
            time_start = datetime.datetime.now()
            string_time_start = time_start.strftime('%Y/%m/%d %H:%M:%S')
            string_time_start_short = time_start.strftime('%y/%m/%d %H:%M:%S')
            if const_meter_temperature == 'DS18B20':
                try:
                    proc = Popen(['cat', '/sys/bus/w1/devices/' + const_DS18B20_address + '/w1_slave'], stdout=PIPE)
                    out, err = proc.communicate()
                    temperature = int(out.decode().split('=')[2])/1000
                    time_current_temperature.set(string_time_start_short)
                    value_current_temperature.set(temperature)
                    time_current_ph.set(string_time_start_short)
                    value_current_ph.set(0)
                    if temperature > const_allowed_max_temperature or temperature < const_allowed_min_temperature:
                        print('Temperature out of range. Resetting temperature power line')
                        #with open('/home/pi/Desktop/Temperature_log_' + time_start.strftime('%Y-%m-%d') + '.txt','a+') as f:
                        #    f.write(string_time_start + '\tTemperature range abnormal: ' + str(temperature) + ',reset\n')
                        GPIO.output(temperature_power_pin, GPIO.LOW)
                        time.sleep(20)
                        GPIO.output(temperature_power_pin, GPIO.HIGH)
                    else:
                        with open('/home/pi/Desktop/Temperature_log_' + time_start.strftime('%Y-%m-%d') + '.txt','a+') as f:
                            f.write(string_time_start + '\tpH=0\tT='  + str(temperature) + '\n')
                except:
                    proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '0']) #set heating off
                    proc.communicate()
                    print(string_time_start + ' Unknown error. Turn off heating and reset temperature power line') # do not write to log because monitoring uses log
                    GPIO.output(temperature_power_pin, GPIO.LOW)
                    time.sleep(20)
                    GPIO.output(temperature_power_pin, GPIO.HIGH)
            elif const_meter_temperature == 'Multiseven':
                try:
                    s = send_multiseven()
                    s = re.sub('Channel: L', '', s)
                    s = re.sub(chr(13), '', s)
                    s = re.sub('\n', '', s)
                    if s != '':
                        t = s.split(' ')
                        flag = 0
                        if float(t[0])>const_allowed_min_pH and float(t[0])<const_allowed_max_pH:
                            time_current_ph.set(string_time_start_short)
                            value_current_ph.set(t[0])
                        else:
                            flag = 1
                        if float(t[2])>const_allowed_min_temperature and float(t[2])<const_allowed_max_temperature:
                            time_current_temperature.set(string_time_start_short)
                            value_current_temperature.set(t[2])
                        else:
                            flag = 1
                        if flag == 0:
                            with open('/home/pi/Desktop/Temperature_log_' + time_start.strftime('%Y-%m-%d') + '.txt','a+') as f:
                                f.write(string_time_start + '\tpH=' + t[0] + '\tT='  + t[2] + '\n')
                except:
                    pass #do not turn off heater, or print or write to log because it happens quite often.
            time.sleep(1)
        button_start_temperature.grid()
        button_stop_temperature.grid_remove()
    thread_temperature = threading.Thread(target=read_temperature)
    thread_temperature.start()
def stop_temperature():
    """stop the thread created by start_temperature"""
    global running_temperature
    running_temperature = False

def start_pressure():
    """start the thread to read from pressure sensor (and/or redox)"""
    """update the current redox/pressure and write to T7Pro_log"""
    """read from pressure control module, and control on/off of valve"""
    def thread_pressure():
        global flag_release_gas_in_progress
        button_start_pressure.grid_remove()
        button_stop_pressure.grid()
        global const_AIN_pressure_reactor
        global const_AIN_pressure_medium
        global const_AIN_redox
        global run_signal # stop reactor when pressure is above critical
        global running_pressure
        running_pressure = True

        global const_meter_pressure
        if const_meter_pressure == 'Labjack':
            global handle
            handle = ljm.openS('T7', 'USB', 'ANY')
            #info = ljm.getHandleInfo(handle)
            """Setup and call eWriteNames to configure AINs on the LabJack."""
            numFrames = 12
            names = ['AIN1_NEGATIVE_CH', 'AIN1_RANGE', 'AIN1_RESOLUTION_INDEX', 'AIN2_NEGATIVE_CH', 'AIN2_RANGE', 'AIN2_RESOLUTION_INDEX', \
                     'AIN0_EF_INDEX', 'AIN0_EF_CONFIG_A', 'AIN0_EF_CONFIG_B', 'AIN0_EF_CONFIG_D', 'AIN0_EF_CONFIG_E', 'AIN0_NEGATIVE_CH', \
                     'AIN0_RANGE', 'AIN0_RESOLUTION_INDEX']
            aValues = [199, 6, 0, 199, 6, 0, 5, 10000, 0, 100000, 10, 199, 1, 0]
            ljm.eWriteNames(handle, numFrames, names, aValues)
        
        elif const_meter_pressure == 'MCP3424':
            bus = smbus.SMBus(1)
            handle_MCP = MCP342x.MCP342x(bus, 0x68, device='MCP3424', channel=1, \
                            gain=1, resolution=18,  continuous_mode=False, scale_factor=1.0, offset=0.0)
            handle_MCP.configure()
        delay = 3 #delay between readings (in sec)
        value_last_pressurev = 0 # initialize last pressure value
        time_last_abnormal_pressurev = datetime.datetime.now() # initialize time for last abnormal pressure
        global abnormal_count
        count = abnormal_count.get()
        global flag_slope_refresh
        global flag_releasing_gas
        flag_releasing_gas.set(False)
        global time_current_pressurev
        global const_critical_pressure
        global const_valve_switch_1
        global const_valve_switch_2
        while running_pressure:
            time_diff = 0
            time_current_pressure = time_current_pressurev.get()
            if const_meter_pressure == 'Labjack':
                """if reading is older than 1 min, close handle and restart handle"""
                if time_current_pressure != '':
                    time_diff = datetime.datetime.now() - datetime.datetime.strptime(time_current_pressure, '%y/%m/%d %H:%M:%S')
                    if time_diff.seconds > 60:
                        ljm.close(handle)
                        handle = ljm.openS('T7', 'USB', 'ANY')
                        ljm.eWriteNames(handle, numFrames, names, aValues)
            """reading the values"""
            try:
                time_start = datetime.datetime.now()
                string_time_start = time_start.strftime('%Y/%m/%d %H:%M:%S')
                string_time_start_short = time_start.strftime('%y/%m/%d %H:%M:%S')
                if const_meter_pressure == 'Labjack':
                    if const_AIN_redox == 'AIN0':
                        ljm.eReadName(handle, 'AIN0_EF_READ_A') #reading AIN0_EF_READ_A is needed to activate AIN0_EF_READ_B
                        redox_voltage = ljm.eReadName(handle, 'AIN0_EF_READ_B')
                    else:
                        redox_voltage = 1 # Assign a fixed value
                    if const_AIN_pressure_reactor == '-':
                        pressure_voltage = 10 # Assign a fixed value
                    else:
                        pressure_voltage = ljm.eReadName(handle, const_AIN_pressure_reactor)
                    if const_AIN_pressure_medium == '-':
                        pressuremedium_voltage = 10 # Assign a fixed value
                    else:
                        pressuremedium_voltage = ljm.eReadName(handle, const_AIN_pressure_medium)
                elif const_meter_pressure == 'MCP3424':
                    redox_voltage = 1 # Assign a fixed value
                    pressure_voltage = handle_MCP.convert_and_read()
                    pressuremedium_voltage = 10 # Assign a fixed value
                """record values to file"""
                with open('/home/pi/Desktop/T7Pro_log_' + time_start.strftime('%Y-%m-%d') + '.txt','a+') as f:    
                    f.write('%s\tRedox=%f\tPressure_V=%f\tPressure_M=%f\n' % (string_time_start, \
                        redox_voltage*1000, pressure_voltage*1000, pressuremedium_voltage*1000))
                """update the display"""
                value_current_pressurev.set(round(pressure_voltage*1000,1))
                time_current_pressurev.set(string_time_start_short)
                value_current_pressuremedium.set(round(pressuremedium_voltage*1000,1))
                time_current_pressuremedium.set(string_time_start_short)
                value_current_redox.set(round(redox_voltage*1000,1))
                time_current_redox.set(string_time_start_short)               
                """record the end pressure after gas release"""
                if flag_slope_refresh.get():
                    with open('/home/pi/Desktop/config/Slope.txt','a+') as f:    
                        f.write('%s\t%f\t' % (string_time_start, pressure_voltage*1000))
                    flag_slope_refresh.set(False)
                """wait for delay in seconds before starting next reading"""
                time.sleep(delay)
            except:
                pass
            """whenever found the current pressure is higher than a defined critical high value,\
                releast gas for 10s, and stop reactor"""
            if pressure_voltage*1000 > const_critical_pressure:
                run_signal = False
                if flag_release_gas_in_progress == False:
                    release_gas(10,True)
            """ If switch_manual_gas_valve is True,do not perform gas release as below,
                but will still release gas when critical pressure is reached (as above)"""
            if switch_manual_gas_valve.get():
                pressure_control.set(False)
            """gas release"""
            if pressure_control.get() and count < 5:
                """ if abnormal increase is more than 5 times during a relatively short period, 
                    do not release gas"""
                try:
                    threshold = input_pressure_threshold.get()
                    release_time = input_release_time.get()
                    time_current_pressure = time_current_pressurev.get()
                    time_now = datetime.datetime.now()
                    #time_now_string = time_now.strftime('%Y/%m/%d %H:%M:%S')
                    #time_now_string_short = time_now.strftime('%y/%m/%d %H:%M:%S')
                    time_diff = time_now - datetime.datetime.strptime(time_current_pressure, '%y/%m/%d %H:%M:%S')
                    current_pressure = float(value_current_pressurev.get())
                    """if reading is older than 1 min, do not relase gas, and keep valve closed"""
                    if time_diff.seconds > 60:
                        value_last_pressurev = current_pressure
                        proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
                        proc.communicate()
                        valve_status.set('closed'  + str(value_reactorV_imbalance.get()))
                        """gas release is activated if threshold is reached, the difference between two adjacent  
                        readings is smaller than 5 mV, and if the gas release is not in progress"""
                    elif current_pressure > threshold \
                        and value_last_pressurev > threshold \
                        and abs(current_pressure - value_last_pressurev) < 5 \
                        and flag_releasing_gas.get() == False:
                        value_last_pressurev = current_pressure
                        """perform gas release"""
                        if flag_release_gas_in_progress:
                            print('release gas in progress')
                        else:
                            release_gas(release_time, False)
                    elif current_pressure - value_last_pressurev > 50: # define abnormal increase as more than 50 mV
                        value_last_pressurev = current_pressure
                        time_diff = time_now - time_last_abnormal_pressurev
                        if time_diff.seconds < 60:
                            count = count +1
                            abnormal_count.set(count)
                            time_last_abnormal_pressurev = time_now
                        else:
                            """if the time difference from the last abnormal reading is more than 60 seconds, reset count to 1"""
                            count = 1
                            abnormal_count.set(count)
                    elif flag_release_gas_in_progress == False:
                        value_last_pressurev = current_pressure
                        proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
                        proc.communicate()
                        valve_status.set('closed ' + str(value_reactorV_imbalance.get()))
                    else:
                        value_last_pressurev = current_pressure
                except:
                    print('Error during pressure control. Close gas valve')
                    proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
                    proc.communicate()
                    valve_status.set('closed ' + str(value_reactorV_imbalance.get()))
            elif count >= 5:
                pressure_control.set(False)
                count = 0
                abnormal_count.set(count)
        if const_meter_pressure == 'Labjack':
            ljm.close(handle) #Close handle
        button_start_pressure.grid()
        button_stop_pressure.grid_remove()
    thread_pressure = threading.Thread(target=thread_pressure)
    thread_pressure.start()
    """make the start button reappear when the loop exits unexpectedly"""
    button_start_pressure.grid()
    button_stop_pressure.grid_remove()
    global running_pressure
    running_pressure = False

def stop_pressure():
    """stop the thread created by start_pressure"""
    global running_pressure
    running_pressure = False
    button_start_pressure.grid()

def release_gas(seconds, critical):
    """release gas for specified seconds, and update the slope
    if critical, release immediately, if not, release when reactor volume is balanced
    """
    def start_release():
        try:
            global flag_release_gas_in_progress
            flag_release_gas_in_progress = True
            global running_reactor
            global const_critical_pressure
            global const_valve_switch_1
            global const_valve_switch_2
            if critical == False: #if release gas due to critical pressure reached (critical == True), do not check syringe movement
                while running_reactor == True \
                      and value_reactorV_imbalance.get() != 0 \
                      and float(value_current_pressurev.get()) < const_critical_pressure: # do not release gas if the flag is True
                    time_now = datetime.datetime.now()
                    valve_status.set('waiting ' + str(value_reactorV_imbalance.get()))
                    time.sleep(3)
            flag_releasing_gas.set(True)
            """write the pressure in slope.txt before gas release"""
            time_now = datetime.datetime.now()
            time_now_string = time_now.strftime('%Y/%m/%d %H:%M:%S')
            #time_now_string_short = time_now.strftime('%y/%m/%d %H:%M:%S')
            try:
                with open('/home/pi/Desktop/config/Slope.txt','a+') as f:
                    time_current_pressurev_short = time_current_pressurev.get()
                    time_current_pressurev_long = datetime.datetime.strptime(time_current_pressurev_short, '%y/%m/%d %H:%M:%S').strftime("%Y/%m/%d %H:%M:%S")
                    f.write('%s\t%f\n' % (time_current_pressurev_long, float(value_current_pressurev.get())))
            except:
                pass
            """start release pressure, send signal twice to make sure it works"""
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '1'])
            proc.communicate()
            valve_status.set('open ' + str(value_reactorV_imbalance.get()))
            time.sleep(0.5)
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '1'])
            proc.communicate()
            time.sleep(int(seconds))
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
            proc.communicate()
            valve_status.set('closed ' + str(value_reactorV_imbalance.get())) 
            time.sleep(0.5)
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
            proc.communicate()
            flag_releasing_gas.set(False)
            flag_slope_refresh.set(True) #this will trigger update of the initial pressure after release (in slope.txt) in pressure thread
            """record release time in T7Pro_log and in last_release"""
            with open('/home/pi/Desktop/T7Pro_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:    
                f.write('%s\tRelease_gas=%f\n' % (time_now_string, seconds))
            with open('/home/pi/Desktop/config/last_release.txt','w') as f:    
                f.write(time_now_string)
            flag_release_gas_in_progress = False
        except:
            """when unexpected error happens, release gas for 20s"""
            time_now = datetime.datetime.now()
            time_now_string = time_now.strftime('%y/%m/%d %H:%M:%S')
            flag_releasing_gas.set(True)
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '1'])
            proc.communicate()
            valve_status.set('open ' + str(value_reactorV_imbalance.get()))
            time.sleep(0.5)
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '1'])
            proc.communicate()
            time.sleep(20)
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
            proc.communicate()
            valve_status.set('closed ' + str(value_reactorV_imbalance.get())) 
            time.sleep(0.5)
            proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
            proc.communicate()
            flag_releasing_gas.set(False)
            with open('/home/pi/Desktop/T7Pro_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:    
                f.write('%s\tRelease_gas=20\n' % (time_now_string))
    thread_release_gas = threading.Thread(target=start_release)
    thread_release_gas.start()

"""Bronkhorst flow meter thread"""
def send_flow(command_text):
    global const_bronkhorst_port
    ser = serial.Serial(port=const_bronkhorst_port,baudrate=38400, \
                        bytesize=8, parity='N', stopbits=1, timeout=1)
    command= command_text + chr(13)
    ser.write(str.encode(command))
    s = ser.read(100)
    ser.close    
    return s

def start_bronkhorst():
    """start the thread to read from Bronkhorst flow meter sensor, update the current flow and total flow and 
        write to Bronkhorst_log"""
    def read_bronkhorst():
        button_start_bronkhorst.grid_remove()
        button_stop_bronkhorst.grid()
        global running_bronkhorst
        running_bronkhorst = True
        while running_bronkhorst:
            time_start = datetime.datetime.now()
            string_time_start = time_start.strftime('%Y/%m/%d %H:%M:%S')
            string_time_start_short = time_start.strftime('%y/%m/%d %H:%M:%S')
            try:
                flow_hex = send_flow(':06800421402140\r\n').decode()
                flow = struct.unpack('!f', bytes.fromhex(flow_hex[11:-2]))[0] #fmeasure: measured value for direct reading (in capunits, max.= capacity)
                value_current_flow.set(round(flow,2))
                time_current_flow.set(string_time_start_short)
                total_hex = send_flow(':06800468416841\r\n').decode()
                total = struct.unpack('!f', bytes.fromhex(total_hex[11:-2]))[0]*1000 #total in ml
                value_current_total_flow.set(round(total,2))
                time_current_total_flow.set(string_time_start_short)
                with open('/home/pi/Desktop/Bronkhorst_log_' + time_start.strftime('%Y-%m-%d') + '.txt','a+') as f:
                    f.write(string_time_start + '\tFlow\t%f\tTotal Flow\t%f\n' %(flow, total))
            except:
                print(string_time_start + '\tBronkhorst error')
                pass
            time.sleep(3)
        button_start_bronkhorst.grid()
        button_stop_bronkhorst.grid_remove()
    thread_bronkhorst = threading.Thread(target=read_bronkhorst)
    thread_bronkhorst.start()
    
def stop_bronkhorst():
    """stop the thread created by start_bronkhorst"""
    global running_bronkhorst
    running_bronkhorst = False 
     
class MainApplication(tk.Frame): #MainApplication is a subclass of tkinter.Frame
    def __init__(self, parent, *args, **kwargs): #this __init__ override the __init__ of the super class
        tk.Frame.__init__(self, parent, *args, **kwargs) #the constructor of the superclass is borrowed
        self.parent = parent #the parent is the root, which is defined in the end of script, root = tk.Tk()
        parent.title('Reactor Control') #change the title of the entire window, i.e., root
        global mainframe
        """set margin in the order of W N E S, unit pixels"""
        mainframe = ttk.Frame(parent, padding='0 0 0 0')  #ttk.Frame is a widget which is a rectangular container for other widgets. It also needs a parent (in this case, the root)
        mainframe.grid(column=0, row=0, sticky=(tk.N, tk.W, tk.E, tk.S))
        min_row_size = 22 #22 fits button height

        #A button to exit the whole window
        #button_exit = ttk.Button(mainframe, text="Quit", command=self.destroy) #the button is located on mainframe, otherwise invisible, but function is to destroy the parent (root)
        #button_exit.grid(column = 0, row = 28)

        for i in range(0, 29):
            mainframe.rowconfigure(i, minsize=min_row_size)
        mainframe.columnconfigure(6, minsize=120) #140 fits datetime yy/mm/dd HH:MM:SS
        mainframe.columnconfigure(7, minsize=50) #50 fits five digits (e.g., 199.6)
        mainframe.columnconfigure(8, minsize=120)
        global running_reactor #required
        running_reactor = False
        global run_signal #required
        run_signal = False
        if const_pump == "Hamilton": #all sections below are based on Hamilton syringe
            self.reactor = Reactor(self)
            self.control = Control(self)
            self.sampling = Sampling(self)
        self.pressurevalve = PressureValve(self)
        self.calculate = Calculate(self)
        self.monitor = Monitor(self)
        self.temperaturecontrol = TemperatureControl(self)
        self.slope = Slope(self)
        global flag_raspberrypi4_or_newer
        if flag_raspberrypi4_or_newer:
            self.plot = Plot(self) # Only activate plot if raspberry pi is 4 or newer
        self.status = Status(self)
        load_program() # load the #default program as set in initialization of class reactor

        
        def kill_window(): #remove teamviewer banner, need to work by installing xdotool
            while True:
                """keep shell=True because it would be complicated to do with shell=False"""
                proc = Popen(['WID=`xdotool search "Sponsored session" | head -1` && xdotool windowkill $WID'], shell=True)
                proc.communicate()
                time.sleep(30)
        thread_killing = threading.Thread(target=kill_window)
        thread_killing.start()
        
class Monitor(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 6
        row_start = 0
        ttk.Label(mainframe, text='Monitoring', font='-weight bold').grid(column=column_start, row=row_start)

        ttk.Label(mainframe, text='pH').grid(column=column_start, row=row_start+1)
        global value_current_ph
        value_current_ph = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_ph).grid(column=column_start+1, row=row_start+1)
        global time_current_ph
        time_current_ph = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_ph).grid(column=column_start+2, row=row_start+1)
        
        ttk.Label(mainframe, text='Temperature').grid(column=column_start, row=row_start+2)
        global value_current_temperature
        value_current_temperature = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_temperature).grid(column=column_start+1, row=row_start+2)
        global time_current_temperature
        time_current_temperature = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_temperature).grid(column=column_start+2, row=row_start+2)

        global running_temperature
        running_temperature = False
        global button_start_temperature
        button_start_temperature = ttk.Button(mainframe, width=7, text='T/pH', command=lambda: start_temperature())
        button_start_temperature.grid(column=column_start+3, row=row_start+1)
        global button_stop_temperature
        button_stop_temperature = ttk.Button(mainframe, width=7, text='Stop T', command=lambda: stop_temperature())
        button_stop_temperature.grid(column=column_start+4, row=row_start+1)
        button_stop_temperature.grid_remove()

        ttk.Label(mainframe, text='Redox').grid(column=column_start, row=row_start+3)
        global value_current_redox
        value_current_redox = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_redox).grid(column=column_start+1, row=row_start+3)
        global time_current_redox
        time_current_redox = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_redox).grid(column=column_start+2, row=row_start+3)
        
        ttk.Label(mainframe, text='Pressure_V').grid(column=column_start, row=row_start+4)
        global value_current_pressurev
        value_current_pressurev = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_pressurev).grid(column=column_start+1, row=row_start+4)
        global time_current_pressurev
        time_current_pressurev = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_pressurev).grid(column=column_start+2, row=row_start+4)

        ttk.Label(mainframe, text='Pressure_M').grid(column=column_start, row=row_start+5)
        global value_current_pressuremedium
        value_current_pressuremedium = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_pressuremedium).grid(column=column_start+1, row=row_start+5)
        global time_current_pressuremedium
        time_current_pressuremedium = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_pressuremedium).grid(column=column_start+2, row=row_start+5)

        global abnormal_count
        abnormal_count = tk.IntVar()
        abnormal_count.set(0)
        ttk.Label(mainframe, textvariable=abnormal_count).grid(column=column_start+3, row=row_start+4)

        global running_pressure
        running_pressure = False
        global button_start_pressure
        button_start_pressure = ttk.Button(mainframe, width=7, text='P/Redox', command=lambda: start_pressure())
        button_start_pressure.grid(column=column_start+3, row=row_start+3)
        global button_stop_pressure
        button_stop_pressure = ttk.Button(mainframe, width=7, text='Stop P', command=lambda: stop_pressure())
        button_stop_pressure.grid(column=column_start+4, row=row_start+3)
        button_stop_pressure.grid_remove()

        ttk.Label(mainframe, text='DO1_amp.').grid(column=column_start, row=row_start+6)
        global value_current_DO1_amplitude
        value_current_DO1_amplitude = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_DO1_amplitude).grid(column=column_start+1, row=row_start+6)
        global time_current_DO1_amplitude
        time_current_DO1_amplitude = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_DO1_amplitude).grid(column=column_start+2, row=row_start+6)

        ttk.Label(mainframe, text='DO1_phase').grid(column=column_start, row=row_start+7)
        global value_current_DO1_phase
        value_current_DO1_phase = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_DO1_phase).grid(column=column_start+1, row=row_start+7)
        global time_current_DO1_phase
        time_current_DO1_phase = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_DO1_phase).grid(column=column_start+2, row=row_start+7)
        
        ttk.Label(mainframe, text='DO1_ppb').grid(column=column_start, row=row_start+8)
        global value_current_DO1_ppb
        value_current_DO1_ppb = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_DO1_ppb).grid(column=column_start+1, row=row_start+8)
        global time_current_DO1_ppb
        time_current_DO1_ppb = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_DO1_ppb).grid(column=column_start+2, row=row_start+8)

        global running_do1
        running_do1 = False
        global button_start_do1
        button_start_do1 = ttk.Button(mainframe, width=7, text='Start DO1', command=lambda: start_do1())
        button_start_do1.grid(column=column_start+3, row=row_start+6)
        global button_stop_do1
        button_stop_do1 = ttk.Button(mainframe, width=7, text='Stop DO1', command=lambda: stop_do1())
        button_stop_do1.grid(column=column_start+4, row=row_start+6)
        button_stop_do1.grid_remove()

        global input_do1_delay
        input_do1_delay = tk.IntVar()
        input_do1_delay.set(600)
        ttk.Label(mainframe, text='Delay (s)').grid(column=column_start+3, row=row_start+7)
        do1_delay_entry = ttk.Entry(mainframe, width=7, textvariable=input_do1_delay, justify='center')
        do1_delay_entry.grid(column=column_start+4, row=row_start+7)
        
        global input_do1_comment
        global const_meter_DO1
        global const_meter_DO1_comment
        input_do1_comment = tk.StringVar()
        input_do1_comment.set(const_meter_DO1 + "-" + const_meter_DO1_comment)
        do1_comment_entry = ttk.Entry(mainframe, width=30, textvariable=input_do1_comment, justify='center')
        do1_comment_entry.grid(column=column_start+3, columnspan=4, row=row_start+8)

        ttk.Label(mainframe, text='DO2_amp.').grid(column=column_start, row=row_start+9)
        global value_current_DO2_amplitude
        value_current_DO2_amplitude = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_DO2_amplitude).grid(column=column_start+1, row=row_start+9)
        global time_current_DO2_amplitude
        time_current_DO2_amplitude = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_DO2_amplitude).grid(column=column_start+2, row=row_start+9)

        ttk.Label(mainframe, text='DO2_phase').grid(column=column_start, row=row_start+10)
        global value_current_DO2_phase
        value_current_DO2_phase = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_DO2_phase).grid(column=column_start+1, row=row_start+10)
        global time_current_DO2_phase
        time_current_DO2_phase = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_DO2_phase).grid(column=column_start+2, row=row_start+10)
        
        ttk.Label(mainframe, text='DO2_ppb').grid(column=column_start, row=row_start+11)
        global value_current_DO2_ppb
        value_current_DO2_ppb = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_DO2_ppb).grid(column=column_start+1, row=row_start+11)
        global time_current_DO2_ppb
        time_current_DO2_ppb = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_DO2_ppb).grid(column=column_start+2, row=row_start+11)

        global running_do2
        running_do2 = False
        global button_start_do2
        button_start_do2 = ttk.Button(mainframe, width=7, text='Start DO2', command=lambda: start_do2())
        button_start_do2.grid(column=column_start+3, row=row_start+9)
        global button_stop_do2
        button_stop_do2 = ttk.Button(mainframe, width=7, text='Stop DO2', command=lambda: stop_do2())
        button_stop_do2.grid(column=column_start+4, row=row_start+9)
        button_stop_do2.grid_remove()

        global input_do2_delay
        input_do2_delay = tk.IntVar()
        input_do2_delay.set(600)
        ttk.Label(mainframe, text='Delay (s)').grid(column=column_start+3, row=row_start+10)
        do2_delay_entry = ttk.Entry(mainframe, width=7, textvariable=input_do2_delay, justify='center')
        do2_delay_entry.grid(column=column_start+4, row=row_start+10)
        
        global input_do2_comment
        global const_meter_DO2
        global const_meter_DO2_comment
        input_do2_comment = tk.StringVar()
        input_do2_comment.set(const_meter_DO2 + "-" + const_meter_DO2_comment)
        do2_comment_entry = ttk.Entry(mainframe, width=30, textvariable=input_do2_comment, justify='center')
        do2_comment_entry.grid(column=column_start+3, columnspan=4, row=row_start+11)

        ttk.Label(mainframe, text='Actual flow').grid(column=column_start, row=row_start+12)
        global value_current_flow
        value_current_flow = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_flow).grid(column=column_start+1, row=row_start+12)
        global time_current_flow
        time_current_flow = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_flow).grid(column=column_start+2, row=row_start+12)

        ttk.Label(mainframe, text='Total flow').grid(column=column_start, row=row_start+13)
        global value_current_total_flow
        value_current_total_flow = tk.StringVar()
        ttk.Label(mainframe, textvariable=value_current_total_flow).grid(column=column_start+1, row=row_start+13)
        global time_current_total_flow
        time_current_total_flow = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_total_flow).grid(column=column_start+2, row=row_start+13)
        
        global running_bronkhorst
        running_bronkhorst = False
        global button_start_bronkhorst
        button_start_bronkhorst = ttk.Button(mainframe, width=7, text='Flow', command=lambda: start_bronkhorst())
        button_start_bronkhorst.grid(column=column_start+3, row=row_start+12)
        global button_stop_bronkhorst
        button_stop_bronkhorst = ttk.Button(mainframe, width=7, text='Stop F', command=lambda: stop_bronkhorst())
        button_stop_bronkhorst.grid(column=column_start+4, row=row_start+12)
        button_stop_bronkhorst.grid_remove()
        
        def start_monitor():
            global monitoring
            global upload
            while True:
                time.sleep(30)
                time_now = datetime.datetime.now()
                """Take a screenshot and upload to dropbox once every 600s"""
                if time.mktime(time_now.timetuple()) % 600 <60: # upload every 600 s
                    screenshot = time_now.strftime('%Y-%m-%d-%H-%M') + '.png'
                    try:
                        if not os.path.exists('/home/pi/Desktop/screenshot/' + time_now.strftime('%Y%m%d') + '/' + screenshot):
                            if not os.path.exists('/home/pi/Desktop/screenshot/' + time_now.strftime('%Y%m%d')):
                                proc = Popen(['mkdir', '/home/pi/Desktop/screenshot/' + time_now.strftime('%Y%m%d')])
                                proc.communicate()
                                time.sleep(1)
                            proc = Popen(["scrot " + screenshot \
                                + " -e 'mv $f ~/Desktop/screenshot/" \
                                + time_now.strftime('%Y%m%d') \
                                + "/' && /home/pi/Dropbox-Uploader/dropbox_uploader.sh \
                                    upload /home/pi/Desktop/screenshot/" \
                                + time_now.strftime('%Y%m%d') + "/" + screenshot \
                                + " /Screenshot/" + time_now.strftime('%Y%m%d') + "/" + screenshot], shell = True)
                            proc.communicate()
                    except:
                        pass
                """Delete screenshot folders that are more than one week old"""
                if os.path.exists('/home/pi/Desktop/screenshot/' + (time_now-datetime.timedelta(days=7)).strftime('%Y%m%d')):
                    proc = Popen(['rm', '-rf', '/home/pi/Desktop/screenshot/' + (time_now-datetime.timedelta(days=7)).strftime('%Y%m%d')])
                    proc.communicate()
                    print('remove screenshots that are older than 7 days')
                """Restart teamviewer if requested"""
                try:
                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                        'download', '/Config/teamviewer.txt', '/home/pi/Desktop/config/teamviewer.txt'])
                    proc.communicate()
                    with open('/home/pi/Desktop/config/teamviewer.txt','r') as f:
                        signal = f.read()
                    if signal[0][0]=='1':
                        print((time.strftime('%Y/%m/%d %H:%M:%S\trestart teamviewer')))
                        proc = Popen(['sudo', 'teamviewer', 'daemon', 'restart']) #sudo needed
                        proc.communicate()
                        with open('/home/pi/Desktop/config/teamviewer.txt','w') as f:
                            f.write('0')
                        proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                            'upload', '/home/pi/Desktop/config/teamviewer.txt', '/Config/teamviewer.txt'])
                        proc.communicate()
                except: #deal with situation where a blank file was downloaded
                    with open('/home/pi/Desktop/config/teamviewer.txt','w') as f:
                            f.write('0')
                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                        'upload', '/home/pi/Desktop/config/teamviewer.txt', '/Config/teamviewer.txt'])
                    proc.communicate()
                    print('unexpected error when reading teamviewer.txt')
                    pass
                """Upload current log files to dropbox if requested"""
                try:
                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                        'download', '/Config/upload.txt', '/home/pi/Desktop/config/upload.txt'])
                    proc.communicate()
                    with open('/home/pi/Desktop/config/upload.txt','r') as f:
                        upload_signal = f.read()
                    if upload_signal[0][0]=='1':
                        print(time_now.strftime('%Y/%m/%d %H:%M:%S\tupload log files'))
                        """upload everything that contains _log_ in the file name on desktop"""
                        os.chdir('/home/pi/Desktop')
                        to_be_uploaded = glob.glob('*_log_*.txt')
                        for child in to_be_uploaded:
                            proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                'upload', '/home/pi/Desktop/' + child, '/Archive/' + child])
                            proc.communicate()
                            print(child + ' uploaded')
                        with open('/home/pi/Desktop/config/upload.txt','w') as f:
                            f.write('0')
                        proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                            'upload', '/home/pi/Desktop/config/upload.txt', '/Config/upload.txt'])
                        proc.communicate()
                except: #deal with situation where a blank file was downloaded
                    with open('/home/pi/Desktop/config/upload.txt','w') as f:
                            f.write('0')
                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                        'upload', '/home/pi/Desktop/config/upload.txt', '/Config/upload.txt'])
                    proc.communicate()
                    print('unexpected error when reading upload.txt')
                    pass
                """perform monitoring if monitor check box is checked and warning_switch.txt is set to 1"""
                if monitoring.get():
                    try:
                        proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                            'download', '/Config/warning_switch.txt', '/home/pi/Desktop/config/warning_switch.txt'])
                        proc.communicate()
                        time.sleep(5)
                        try:
                            with open('/home/pi/Desktop/config/warning_switch.txt','r') as f:
                                warning = f.read()
                        except:
                            time_now = datetime.datetime.now()
                            string_time_now = time_now.strftime('%Y/%m/%d %H:%M:%S')
                            #string_time_now_short = time_now.strftime('%y/%m/%d %H:%M:%S')
                            print(string_time_now + '\terror in opening warning_switch.txt.')
                        if warning[0][0]=='1':
                            time_now = datetime.datetime.now()
                            temperature_log = '/home/pi/Desktop/Temperature_log_' + time_now.strftime('%Y-%m-%d') + '.txt'
                            t7pro_log = '/home/pi/Desktop/T7Pro_log_' + time_now.strftime('%Y-%m-%d') + '.txt'
                            if os.path.exists(temperature_log) and os.path.exists(t7pro_log):
                                f = open('/home/pi/Desktop/Temperature_log_' + time_now.strftime('%Y-%m-%d') + '.txt', 'rb')
                                Temperature_1 = tail(f, 1)
                                Temperature_2 = Temperature_1.split('\t')
                                time_temperature = Temperature_2[0]
                                try:
                                    temperature = Temperature_2[2].replace('T=', '')
                                except:
                                    temperature = "There is no valid temperature input in the last line of the log file."
                                
                                f = open('/home/pi/Desktop/T7Pro_log_' + time_now.strftime('%Y-%m-%d') + '.txt', 'rb')
                                pressure_1 = tail(f, 1)

                                pressure_2 = pressure_1.split('\t') #[0] time, [1] redox, [2] pressure of reactor, [3] 2nd pressure (medium or other)
                                if 'Release_gas' in pressure_2[1]:
                                    time_pressure = pressure_2[0]
                                    pressure = 0
                                else:
                                    pressure_3 = pressure_2[2].split('=')
                                    time_pressure = pressure_2[0]
                                    pressure = pressure_3[1]                          

                                global time_last_monitor
                                time_last_monitor.set(time_now.strftime('%m/%d %H:%M'))
                                global const_critical_pressure
                                time_diff_pH = datetime.datetime.now() - datetime.datetime.strptime(time_temperature, '%Y/%m/%d %H:%M:%S')
                                time_diff_pressure = datetime.datetime.now() - datetime.datetime.strptime(time_pressure, '%Y/%m/%d %H:%M:%S')
                                if time_diff_pH.seconds > 60*20 or time_diff_pressure.seconds > 60*20:
                                    with open('/home/pi/Desktop/config/warning_message.txt','w') as f:
                                        f.write('The temperature or pressure log file is not updated for at least 20 min.')
                                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                        'upload', '/home/pi/Desktop/config/warning_message.txt', '/warning/' \
                                        + time_now.strftime('%Y_%m_%d_%H_%M_%S') + '.txt'])
                                    proc.communicate()
                                    with open('/home/pi/Desktop/config/warning_switch.txt','w') as f:
                                        f.write('0')
                                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                        'upload', '/home/pi/Desktop/config/warning_switch.txt', '/Config/warning_switch.txt'])
                                    proc.communicate()
                                    monitoring.set(False)
                                elif float(temperature)>(input_temperature.get() + 1.0) or float(pressure) > const_critical_pressure:
                                    with open('/home/pi/Desktop/config/warning_message.txt','w') as f:
                                        f.write(time_temperature + ': ' + temperature + '\t' + time_pressure + ': ' + pressure)
                                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                        'upload', '/home/pi/Desktop/config/warning_message.txt', '/warning/' \
                                        + time_now.strftime('%Y_%m_%d_%H_%M_%S') + '.txt'])
                                    proc.communicate()
                                    with open('/home/pi/Desktop/config/warning_switch.txt','w') as f:
                                        f.write('0')
                                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                        'upload', '/home/pi/Desktop/config/warning_switch.txt', '/Config/warning_switch.txt'])
                                    proc.communicate()
                                    monitoring.set(False)
                                
                            elif time_now.time() > datetime.time(hour=0, minute=20):
                                with open('/home/pi/Desktop/config/warning_message.txt','w') as f:
                                    f.write('One of the two log files hasn not been created after 20 min.')
                                proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                    'upload', '/home/pi/Desktop/config/warning_message.txt', '/warning/' \
                                    + time_now.strftime('%Y_%m_%d_%H_%M_%S') + '.txt'])
                                proc.communicate()
                                with open('/home/pi/Desktop/config/warning_switch.txt','w') as f:
                                    f.write('0')
                                proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                    'upload', '/home/pi/Desktop/config/warning_switch.txt', '/Config/warning_switch.txt'])
                                proc.communicate()
                                monitoring.set(False)
                    except Exception as error:
                        time_now = datetime.datetime.now()
                        string_time_now = time_now.strftime('%Y/%m/%d %H:%M:%S')
                        print(string_time_now + '\tmonitoring encountered an error:' + str(repr(error)))
                        monitoring.set(False)
                """perform upload of all log files before today, if upload checkbox is checked"""  
                if upload.get():    
                    try:
                        time_now = datetime.datetime.now()
                        if time_now.time() > datetime.time(minute=10): #perform the upload only after 0:10:00
                            """upload everything that contains _log_ in the file name on desktop"""
                            os.chdir('/home/pi/Desktop')
                            to_be_uploaded = glob.glob('*_log_*.txt')
                            for child in to_be_uploaded:
                                temp = child.split('_')
                                file_date = temp[2][:-4]
                                if file_date != time_now.strftime('%Y-%m-%d'): #skip today's log file
                                    newdir = '/home/pi/Desktop/log/' + file_date
                                    if not os.path.exists(newdir):
                                        proc = Popen(['mkdir', newdir])
                                        proc.communicate()
                                        print('create log folder for today')
                                    proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                                        'upload', '/home/pi/Desktop/' + child, '/Archive/' + file_date + '/' + child])
                                    proc.communicate()
                                    proc = Popen(['mv', '/home/pi/Desktop/' + child, newdir + '/'])
                                    proc.communicate()
                                    global time_last_upload
                                    time_last_upload.set(time_now.strftime('%m/%d %H:%M'))
                    except:
                        pass
        thread_monitor = threading.Thread(target=start_monitor)
        thread_monitor.start()
    
class Reactor(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 0
        row_start = 0
        ttk.Label(mainframe, text='Reactor', font='-weight bold').grid(column=column_start, row=row_start)
        ttk.Label(mainframe, text='Set').grid(column=column_start+1, row=row_start)
        ttk.Label(mainframe, text='Current').grid(column=column_start+2, row=row_start)
        global input_cycletime
        input_cycletime = tk.StringVar()
        global value_current_cycletime
        value_current_cycletime = tk.StringVar()
        ttk.Label(mainframe, text='Cycle time').grid(column=column_start, row=row_start+1)
        cycletime_entry = ttk.Entry(mainframe, width=13, textvariable=input_cycletime, justify='center')
        cycletime_entry.grid(column=column_start+1, row=row_start+1)
        ttk.Entry(mainframe, width=13, textvariable=value_current_cycletime, justify='center').grid(column=column_start+2, row=row_start+1)
        global input_cycle
        input_cycle = tk.IntVar()
        global value_current_cycle
        value_current_cycle = tk.IntVar()
        ttk.Label(mainframe, text='Cycle').grid(column=column_start, row=row_start+2)
        cycle_entry = ttk.Entry(mainframe, width=13, textvariable=input_cycle, justify='center')
        cycle_entry.grid(column=column_start+1, row=row_start+2)
        ttk.Entry(mainframe, width=13, textvariable=value_current_cycle, justify='center').grid(column=column_start+2, row=row_start+2)

        global value_reactorV_imbalance
        value_reactorV_imbalance = tk.DoubleVar()
        value_reactorV_imbalance.set(0) #when initialized, reactor volume is balanced (or to say, same as the original volume)

        global current_entry_step1
        global current_entry_step2
        global current_entry_step3
        global current_entry_step4
        global current_entry_step5
        global current_entry_step6
        global current_entry_step7
        global current_entry_step8
        global current_entry_step9
        global current_entry_step10
        global current_entry_step11
        global current_entry_step12
        global current_entry_step13
        global current_entry_step14
        global current_entry_step15
        global current_entry_step16
        global current_entry_step17
        global current_entry_step18
        global current_entry_step19
        global current_entry_step20
        global current_entry_step21
        global current_entry_step22
        global current_entry_step23
        global current_entry_step24
        global current_entry_step25
        
        global input_step1
        input_step1 = tk.StringVar()
        global value_current_step1
        value_current_step1 = tk.StringVar()
        ttk.Label(mainframe, text='Step 1').grid(column=column_start, row=row_start+3)
        ttk.Entry(mainframe, width=13, textvariable=input_step1, justify='center').grid(column=column_start+1, row=row_start+3)
        current_entry_step1 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step1)
        current_entry_step1.grid(column=column_start+2, row=row_start+3)       

        global input_step2
        input_step2 = tk.StringVar()
        global value_current_step2
        value_current_step2 = tk.StringVar()
        ttk.Label(mainframe, text='Step 2').grid(column=column_start, row=row_start+4)
        ttk.Entry(mainframe, width=13, textvariable=input_step2, justify='center').grid(column=column_start+1, row=row_start+4)
        current_entry_step2 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step2)
        current_entry_step2.grid(column=column_start+2, row=row_start+4)       

        global input_step3
        input_step3 = tk.StringVar()
        global value_current_step3
        value_current_step3 = tk.StringVar()
        ttk.Label(mainframe, text='Step 3').grid(column=column_start, row=row_start+5)
        ttk.Entry(mainframe, width=13, textvariable=input_step3, justify='center').grid(column=column_start+1, row=row_start+5)
        current_entry_step3 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step3)
        current_entry_step3.grid(column=column_start+2, row=row_start+5)        

        global input_step4
        input_step4 = tk.StringVar()
        global value_current_step4
        value_current_step4 = tk.StringVar()
        ttk.Label(mainframe, text='Step 4').grid(column=column_start, row=row_start+6)
        ttk.Entry(mainframe, width=13, textvariable=input_step4, justify='center').grid(column=column_start+1, row=row_start+6)
        current_entry_step4 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step4)
        current_entry_step4.grid(column=column_start+2, row=row_start+6)       

        global input_step5
        input_step5 = tk.StringVar()
        global value_current_step5
        value_current_step5 = tk.StringVar()
        ttk.Label(mainframe, text='Step 5').grid(column=column_start, row=row_start+7)
        ttk.Entry(mainframe, width=13, textvariable=input_step5, justify='center').grid(column=column_start+1, row=row_start+7)
        current_entry_step5 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step5)
        current_entry_step5.grid(column=column_start+2, row=row_start+7)       

        global input_step6
        input_step6 = tk.StringVar()
        global value_current_step6
        value_current_step6 = tk.StringVar()
        ttk.Label(mainframe, text='Step 6').grid(column=column_start, row=row_start+8)
        ttk.Entry(mainframe, width=13, textvariable=input_step6, justify='center').grid(column=column_start+1, row=row_start+8)
        current_entry_step6 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step6)
        current_entry_step6.grid(column=column_start+2, row=row_start+8)

        global input_step7
        input_step7 = tk.StringVar()
        global value_current_step7
        value_current_step7 = tk.StringVar()
        ttk.Label(mainframe, text='Step 7').grid(column=column_start, row=row_start+9)
        ttk.Entry(mainframe, width=13, textvariable=input_step7, justify='center').grid(column=column_start+1, row=row_start+9)
        current_entry_step7 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step7)
        current_entry_step7.grid(column=column_start+2, row=row_start+9)       

        global input_step8
        input_step8 = tk.StringVar()
        global value_current_step8
        value_current_step8 = tk.StringVar()
        ttk.Label(mainframe, text='Step 8').grid(column=column_start, row=row_start+10)
        ttk.Entry(mainframe, width=13, textvariable=input_step8, justify='center').grid(column=column_start+1, row=row_start+10)
        current_entry_step8 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step8)
        current_entry_step8.grid(column=column_start+2, row=row_start+10)       

        global input_step9
        input_step9 = tk.StringVar()
        global value_current_step9
        value_current_step9 = tk.StringVar()
        ttk.Label(mainframe, text='Step 9').grid(column=column_start, row=row_start+11)
        ttk.Entry(mainframe, width=13, textvariable=input_step9, justify='center').grid(column=column_start+1, row=row_start+11)
        current_entry_step9 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step9)
        current_entry_step9.grid(column=column_start+2, row=row_start+11)      

        global input_step10
        input_step10 = tk.StringVar()
        global value_current_step10
        value_current_step10 = tk.StringVar()
        ttk.Label(mainframe, text='Step 10').grid(column=column_start, row=row_start+12)
        ttk.Entry(mainframe, width=13, textvariable=input_step10, justify='center').grid(column=column_start+1, row=row_start+12)
        current_entry_step10 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step10)
        current_entry_step10.grid(column=column_start+2, row=row_start+12)       

        global input_step11
        input_step11 = tk.StringVar()
        global value_current_step11
        value_current_step11 = tk.StringVar()
        ttk.Label(mainframe, text='Step 11').grid(column=column_start, row=row_start+13)
        ttk.Entry(mainframe, width=13, textvariable=input_step11, justify='center').grid(column=column_start+1, row=row_start+13)
        current_entry_step11 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step11)
        current_entry_step11.grid(column=column_start+2, row=row_start+13)       

        global input_step12
        input_step12 = tk.StringVar()
        global value_current_step12
        value_current_step12 = tk.StringVar()
        ttk.Label(mainframe, text='Step 12').grid(column=column_start, row=row_start+14)
        ttk.Entry(mainframe, width=13, textvariable=input_step12, justify='center').grid(column=column_start+1, row=row_start+14)
        current_entry_step12 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step12)
        current_entry_step12.grid(column=column_start+2, row=row_start+14)     

        global input_step13
        input_step13 = tk.StringVar()
        global value_current_step13
        value_current_step13 = tk.StringVar()
        ttk.Label(mainframe, text='Step 13').grid(column=column_start, row=row_start+15)
        ttk.Entry(mainframe, width=13, textvariable=input_step13, justify='center').grid(column=column_start+1, row=row_start+15)
        current_entry_step13 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step13)
        current_entry_step13.grid(column=column_start+2, row=row_start+15)       

        global input_step14
        input_step14 = tk.StringVar()
        global value_current_step14
        value_current_step14 = tk.StringVar()
        ttk.Label(mainframe, text='Step 14').grid(column=column_start, row=row_start+16)
        ttk.Entry(mainframe, width=13, textvariable=input_step14, justify='center').grid(column=column_start+1, row=row_start+16)
        current_entry_step14 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step14)
        current_entry_step14.grid(column=column_start+2, row=row_start+16)        

        global input_step15
        input_step15 = tk.StringVar()
        global value_current_step15
        value_current_step15 = tk.StringVar()
        ttk.Label(mainframe, text='Step 15').grid(column=column_start, row=row_start+17)
        ttk.Entry(mainframe, width=13, textvariable=input_step15, justify='center').grid(column=column_start+1, row=row_start+17)
        current_entry_step15 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step15)
        current_entry_step15.grid(column=column_start+2, row=row_start+17)       

        global input_step16
        input_step16 = tk.StringVar()
        global value_current_step16
        value_current_step16 = tk.StringVar()
        ttk.Label(mainframe, text='Step 16').grid(column=column_start, row=row_start+18)
        ttk.Entry(mainframe, width=13, textvariable=input_step16, justify='center').grid(column=column_start+1, row=row_start+18)
        current_entry_step16 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step16)
        current_entry_step16.grid(column=column_start+2, row=row_start+18)       

        global input_step17
        input_step17 = tk.StringVar()
        global value_current_step17
        value_current_step17 = tk.StringVar()
        ttk.Label(mainframe, text='Step 17').grid(column=column_start, row=row_start+19)
        ttk.Entry(mainframe, width=13, textvariable=input_step17, justify='center').grid(column=column_start+1, row=row_start+19)
        current_entry_step17 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step17)
        current_entry_step17.grid(column=column_start+2, row=row_start+19)       

        global input_step18
        input_step18 = tk.StringVar()
        global value_current_step18
        value_current_step18 = tk.StringVar()
        ttk.Label(mainframe, text='Step 18').grid(column=column_start, row=row_start+20)
        ttk.Entry(mainframe, width=13, textvariable=input_step18, justify='center').grid(column=column_start+1, row=row_start+20)
        current_entry_step18 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step18)
        current_entry_step18.grid(column=column_start+2, row=row_start+20)

        global input_step19
        input_step19 = tk.StringVar()
        global value_current_step19
        value_current_step19 = tk.StringVar()
        ttk.Label(mainframe, text='Step 19').grid(column=column_start, row=row_start+21)
        ttk.Entry(mainframe, width=13, textvariable=input_step19, justify='center').grid(column=column_start+1, row=row_start+21)
        current_entry_step19 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step19)
        current_entry_step19.grid(column=column_start+2, row=row_start+21)  

        global input_step20
        input_step20 = tk.StringVar()
        global value_current_step20
        value_current_step20 = tk.StringVar()
        ttk.Label(mainframe, text='Step 20').grid(column=column_start, row=row_start+22)
        ttk.Entry(mainframe, width=13, textvariable=input_step20, justify='center').grid(column=column_start+1, row=row_start+22)
        current_entry_step20 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step20)
        current_entry_step20.grid(column=column_start+2, row=row_start+22)

        global input_step21
        input_step21 = tk.StringVar()
        global value_current_step21
        value_current_step21 = tk.StringVar()
        ttk.Label(mainframe, text='Step 21').grid(column=column_start, row=row_start+23)
        ttk.Entry(mainframe, width=13, textvariable=input_step21, justify='center').grid(column=column_start+1, row=row_start+23)
        current_entry_step21 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step21)
        current_entry_step21.grid(column=column_start+2, row=row_start+23)

        global input_step22
        input_step22 = tk.StringVar()
        global value_current_step22
        value_current_step22 = tk.StringVar()
        ttk.Label(mainframe, text='Step 22').grid(column=column_start, row=row_start+24)
        ttk.Entry(mainframe, width=13, textvariable=input_step22, justify='center').grid(column=column_start+1, row=row_start+24)
        current_entry_step22 = ttk.Entry(mainframe, width=13, justify='center', textvariable=value_current_step22)
        current_entry_step22.grid(column=column_start+2, row=row_start+24)

        global lastcycle
        lastcycle = tk.StringVar()
        ttk.Label(mainframe, text='Last cycle').grid(column=column_start, row=row_start+25)
        ttk.Label(mainframe, textvariable=lastcycle).grid(column=column_start+1, columnspan=2, row=row_start+25)

        global nextcycle
        nextcycle = tk.StringVar()
        ttk.Label(mainframe, text='Next cycle').grid(column=column_start, row=row_start+26)
        ttk.Label(mainframe, textvariable=nextcycle).grid(column=column_start+1, columnspan=2, row=row_start+26)

        ttk.Label(mainframe, text='Program').grid(column=column_start, row=row_start+27)
        global input_program
        global const_default_program
        input_program = tk.StringVar()
        input_program.set('#' + const_default_program)
        program_entry = ttk.Entry(mainframe, width=10, textvariable=input_program, justify='center')
        program_entry.grid(column=column_start+1, row=row_start+27)
        global button_load_program
        button_load_program = ttk.Button(mainframe, text='Load', width=10, command=lambda: load_program())
        button_load_program.grid(column=column_start+2, row=row_start+27)

        global label_running
        label_running = ttk.Label(mainframe, width=10, text='Running')
        label_running.grid(column=column_start+1, row=row_start+28)
        label_running.grid_remove()
        global label_stopping
        label_stopping = ttk.Label(mainframe, width=10, text='Stopping')
        label_stopping.grid(column=column_start+2, row=row_start+28)
        label_stopping.grid_remove()
        global button_start_reactor
        button_start_reactor = ttk.Button(mainframe, text='Start', width=10, command=lambda: start_reactor())
        button_start_reactor.grid(column=column_start+1, row=row_start+28)
        global button_stop_reactor
        button_stop_reactor = ttk.Button(mainframe, text='Stop', width=10, command=lambda: stop_reactor())
        button_stop_reactor.grid(column=column_start+2, row=row_start+28)
        button_stop_reactor.grid_remove()
        cycletime_entry.focus()

class Calculate(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 3
        row_start = 20
        ttk.Label(mainframe, text='Calculate', font='-weight bold').grid(column=column_start, row=row_start)
        
        ttk.Label(mainframe, text='Nitrite (mM)').grid(column=column_start, row=row_start+1)
        global nitrite_in_mM_in_final_medium
        nitrite_in_mM_in_final_medium = tk.DoubleVar()
        global const_nitrite_in_mM_in_final_medium
        nitrite_in_mM_in_final_medium.set(const_nitrite_in_mM_in_final_medium)
        ttk.Label(mainframe, textvariable=nitrite_in_mM_in_final_medium).grid(column=column_start, row=row_start+2)
        
        ttk.Label(mainframe, text='V(uL) per cycle').grid(column=column_start+1, row=row_start+1)
        global input_flow
        input_flow = tk.DoubleVar()
        flow_entry = ttk.Entry(mainframe, width=7, textvariable=input_flow, justify='center')
        flow_entry.grid(column=column_start+1, row=row_start+2)

        ttk.Label(mainframe, text='Nitrite/N2 ratio').grid(column=column_start, row=row_start+3)
        global nitrite_to_N2_ratio
        nitrite_to_N2_ratio = tk.DoubleVar()
        global const_nitrite_to_N2_ratio
        nitrite_to_N2_ratio.set(const_nitrite_to_N2_ratio)
        ttk.Label(mainframe, textvariable=nitrite_to_N2_ratio).grid(column=column_start, row=row_start+4)

        ttk.Label(mainframe, text='mV / mL gas').grid(column=column_start+1, row=row_start+3)
        global mV_per_mL_gas #e.g., how many mV equals to 1ml gas production
        mV_per_mL_gas = tk.DoubleVar()
        global const_mV_per_mL_gas
        mV_per_mL_gas.set(const_mV_per_mL_gas)
        ttk.Label(mainframe, textvariable=mV_per_mL_gas).grid(column=column_start+1, row=row_start+4)

        ttk.Label(mainframe, text='Gas T.').grid(column=column_start+2, row=row_start+3)
        global gas_temperature_for_calculation
        gas_temperature_for_calculation = tk.DoubleVar()
        global const_gas_temperature_for_calculation
        gas_temperature_for_calculation.set(const_gas_temperature_for_calculation)
        ttk.Label(mainframe, textvariable=gas_temperature_for_calculation).grid(column=column_start+2, row=row_start+4)

        ttk.Label(mainframe, text='Reactor V. (mL)').grid(column=column_start, row=row_start+5)
        global reactor_size
        reactor_size = tk.IntVar()
        global const_reactor_size
        reactor_size.set(const_reactor_size)
        ttk.Label(mainframe, textvariable=reactor_size).grid(column=column_start, row=row_start+6)
        
        ttk.Label(mainframe, text='Flow (mL/d)').grid(column=column_start+1, row=row_start+5)
        global flow_per_day
        flow_per_day = tk.DoubleVar()
        ttk.Label(mainframe, textvariable=flow_per_day).grid(column=column_start+1, row=row_start+6)

        ttk.Label(mainframe, text='Gas (mL/d)').grid(column=column_start+2, row=row_start+5)
        global gas_per_day
        gas_per_day = tk.DoubleVar()
        ttk.Label(mainframe, textvariable=gas_per_day).grid(column=column_start+2, row=row_start+6)

        ttk.Label(mainframe, text='Doubling time (d)').grid(column=column_start, row=row_start+7)
        global doubling_time
        doubling_time = tk.DoubleVar()
        ttk.Label(mainframe, textvariable=doubling_time).grid(column=column_start, row=row_start+8)

        ttk.Label(mainframe, text='Slope (mV/d)').grid(column=column_start+1, row=row_start+7)
        global slope_calculated #calcualted slope (mV/d)
        slope_calculated = tk.DoubleVar()
        ttk.Label(mainframe, textvariable=slope_calculated).grid(column=column_start+1, row=row_start+8)

        global update_in_progress
        update_in_progress = False
        def update_calc(*args):
            global update_in_progress
            global const_reactor_size
            global const_nitrite_in_mM_in_final_medium
            global const_mV_per_mL_gas
            global const_gas_temperature_for_calculation
            global const_nitrite_to_N2_ratio
            if update_in_progress:
                return
            try:
                cycletime_int = int(input_cycletime.get())
                nitrite_float = const_nitrite_in_mM_in_final_medium
                flow_float = input_flow.get()
                reactorV_float = const_reactor_size
                slope_gas_ratio = const_mV_per_mL_gas
                gas_temperature_float = const_gas_temperature_for_calculation
                nitrite_n2_ratio_float = const_nitrite_to_N2_ratio
                new_flow_per_day = round(flow_float/1000*86400/cycletime_int,1)
                new_gas_per_day = round(86400/cycletime_int*0.001 \
                    *flow_float*0.001*nitrite_float/nitrite_n2_ratio_float*22.4*(273+gas_temperature_float)/273,1)                       
                new_doubling_time = round(log(2)*reactorV_float/(86400/cycletime_int*flow_float*0.001),2)
                new_slope_calculated = round(slope_gas_ratio * new_gas_per_day,0)
                update_in_progress = True
                flow_per_day.set(new_flow_per_day)
                gas_per_day.set(new_gas_per_day)
                doubling_time.set(new_doubling_time)
                slope_calculated.set(new_slope_calculated)
                update_in_progress = False
            except:
                pass
        update_calc()
        try:
            input_cycletime.trace_add('write', update_calc)
            input_flow.trace_add('write', update_calc)
        except: #in case python version is lower than 3.6
            input_cycletime.trace_variable('w', update_calc)
            input_flow.trace_variable('w', update_calc)            
         
class Control(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 3
        row_start = 4
        ttk.Label(mainframe, text='Control', font='-weight bold').grid(column=column_start, row=row_start)
        global input_move_syringe_position
        global input_move_syringe_speed
        input_move_syringe_position = tk.IntVar()
        input_move_syringe_position.set(3000)
        input_move_syringe_speed = tk.IntVar()
        input_move_syringe_speed.set(500)
        ttk.Label(mainframe, text='Syringe position').grid(column=column_start, row=row_start+1)
        ttk.Entry(mainframe, width=7, textvariable=input_move_syringe_position, justify='center').grid(column=column_start+1, row=row_start+1)
        ttk.Label(mainframe, text='Syringe speed').grid(column=column_start, row=row_start+2)
        ttk.Entry(mainframe, width=7, textvariable=input_move_syringe_speed, justify='center').grid(column=column_start+1, row=row_start+2)
        global button_move_syringe
        button_move_syringe = ttk.Button(mainframe, text='Move syringe', width=10, \
            command=lambda: move_syringe_standalone(int(input_move_syringe_speed.get()),int(input_move_syringe_position.get())))
        button_move_syringe.grid(column=column_start+2, row=row_start+1, rowspan=2)
        ttk.Button(mainframe, text='Init. PSD4', width=10, command=lambda: initialize_standalone()).grid(column=column_start+2, row=row_start)
        
        global input_move_valve_position
        input_move_valve_position = tk.IntVar()
        input_move_valve_position.set(1)
        ttk.Label(mainframe, text='Valve').grid(column=column_start, row=row_start+3)
        ttk.Entry(mainframe, width=7, textvariable=input_move_valve_position, justify='center').grid(column=column_start+1, row=row_start+3)
        global button_move_valve
        button_move_valve = ttk.Button(mainframe, text='Turn valve', width=10, \
            command=lambda: move_valve_standalone(int(input_move_valve_position.get())))
        button_move_valve.grid(column=column_start+2, row=row_start+3)

        global value_current_valve
        value_current_valve = tk.IntVar()
        value_current_valve.set(-1) #initialize with a negative value before receiving the actual value from the pump
        global value_current_syringe
        value_current_syringe = tk.IntVar()
        value_current_syringe.set(-1) #initialize with a negative value before receiving the actual value from the pump
        ttk.Label(mainframe, text='Current valve').grid(column=column_start, row=row_start+4)
        ttk.Label(mainframe, textvariable=value_current_valve).grid(column=column_start+1, row=row_start+4)
        ttk.Label(mainframe, text='Current syringe').grid(column=column_start, row=row_start+5)
        ttk.Label(mainframe, textvariable=value_current_syringe).grid(column=column_start+1, row=row_start+5)
        try:
            send_hamilton('h30001') #change to H mode to check syringe and valve
            self.check_status(self)
        except:
            pass
        global button_check_status
        button_check_status = ttk.Button(mainframe, text='Update', width=10, command=lambda: self.check_status(self))
        button_check_status.grid(column=column_start+2, row=row_start+4)

        global const_syringe_size
        global current_syringe_size
        current_syringe_size = tk.IntVar()
        current_syringe_size.set(const_syringe_size)
        ttk.Label(mainframe, text='Syringe size').grid(column=column_start, row=row_start+6)
        ttk.Label(mainframe, textvariable=current_syringe_size).grid(column=column_start+1, row=row_start+6)
 
        ttk.Label(mainframe, text='Monitoring').grid(column=column_start, row=row_start+7)
        global monitoring
        monitoring = tk.BooleanVar()
        monitoring.set(False)
        monitoring.set(True)
        with open('/home/pi/Desktop/config/warning_switch.txt','w') as f:
            f.write('1')
        proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
            'upload', '/home/pi/Desktop/config/warning_switch.txt', '/Config/warning_switch.txt'])
        proc.communicate()
        ttk.Checkbutton(mainframe, text='on/off', variable=monitoring).grid(column=column_start+1, row=row_start+7)
        def update_monitoring_switch(*args):
            if monitoring.get() == True:
                with open('/home/pi/Desktop/config/warning_switch.txt','w') as f:
                    f.write('1')
                proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                    'upload', '/home/pi/Desktop/config/warning_switch.txt', '/Config/warning_switch.txt'])
                proc.communicate()
            else:
                with open('/home/pi/Desktop/config/warning_switch.txt','w') as f:
                    f.write('0')
                proc = Popen(['/home/pi/Dropbox-Uploader/dropbox_uploader.sh', \
                    'upload', '/home/pi/Desktop/config/warning_switch.txt', '/Config/warning_switch.txt'])
                proc.communicate()
        try:
            monitoring.trace_add('write', update_monitoring_switch)
        except: #in case python version is lower than 3.6
            monitoring.trace_variable('w', update_monitoring_switch)
        global time_last_monitor
        time_last_monitor = tk.StringVar()
        time_last_monitor.set('')
        ttk.Label(mainframe, textvariable=time_last_monitor).grid(column=column_start+2, row=row_start+7)

        ttk.Label(mainframe, text='Upload').grid(column=column_start, row=row_start+8)
        global upload
        upload = tk.BooleanVar()
        upload.set(True)
        ttk.Checkbutton(mainframe, text='on/off', variable=upload).grid(column=column_start+1, row=row_start+8)
        global time_last_upload
        time_last_upload = tk.StringVar()
        time_last_upload.set('')
        ttk.Label(mainframe, textvariable=time_last_upload).grid(column=column_start+2, row=row_start+8)

        ttk.Label(mainframe, text='Stirring').grid(column=column_start, row=row_start+9)
        global stirring
        global flag_stirring
        flag_stirring = False
        global const_stirring_switch_1
        global const_stirring_switch_2
        stirring = tk.BooleanVar()
        stirring.set(False)
        proc = Popen(['/home/pi/raspberry-remote/send', const_stirring_switch_1, const_stirring_switch_2, '0'])
        proc.communicate()
        ttk.Checkbutton(mainframe, text='on/off', variable=stirring).grid(column=column_start+1, row=row_start+9)
        global stirring_interval
        global const_stirring_interval
        stirring_interval = tk.IntVar()
        stirring_interval.set(const_stirring_interval) # interval in seconds
        stirring_interval_entry = ttk.Entry(mainframe, width=7, textvariable=str(stirring_interval), justify='center')
        stirring_interval_entry.grid(column=column_start+2, row=row_start+9)
        
        def run_stirring():
            def start_stirring():
                global flag_stirring
                start = datetime.datetime.now()
                try:
                    interval = int(stirring_interval.get())
                    if interval < 600:
                        stirring.set(False)
                        mbox.showinfo('Error', 'Stirring interval has to be at least 600 s.')
                        return 0
                except:
                    stirring.set(False)
                    mbox.showinfo('Error', 'Error with stirring interval format (must be integers). Quit stirring')
                    return 0
                end = start + datetime.timedelta(seconds=int(interval))       
                while flag_stirring == True:
                    time_now = datetime.datetime.now()
                    time_now_string = time_now.strftime('%Y/%m/%d %H:%M:%S')
                    proc = Popen(['/home/pi/raspberry-remote/send', const_stirring_switch_1, const_stirring_switch_2, '1'])
                    proc.communicate()
                    time.sleep(60) #stirring is done for a fixed time of 60 s

                    with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:    
                        f.write('%s\tStirring for 60 seconds\n' % (time_now_string))

                    while datetime.datetime.now() < end and flag_stirring == True:
                        try:
                            interval = int(stirring_interval.get())
                            if interval < 600:
                                stirring.set(False)
                                mbox.showinfo('Error', 'Stirring interval has to be at least 600 s.')
                                return 0
                        except:
                            stirring.set(False)
                            mbox.showinfo('Error', 'Error with stirring interval format (must be integers). Quit stirring')
                            return 0
                        end = start + datetime.timedelta(seconds=int(interval))
                        #Keep sending off signal during the waiting period
                        proc = Popen(['/home/pi/raspberry-remote/send', const_stirring_switch_1, const_stirring_switch_2, '0'])
                        proc.communicate()
                        time.sleep(3)
                        if flag_stirring == False:
                            print("Stirring was stopped.")
                    start = end
                    end = start + datetime.timedelta(seconds=int(interval))
                #Send signal 3 times to make sure the stirrer is switched off.
                proc = Popen(['/home/pi/raspberry-remote/send', const_stirring_switch_1, const_stirring_switch_2, '0'])
                proc.communicate()
                time.sleep(3)
                proc = Popen(['/home/pi/raspberry-remote/send', const_stirring_switch_1, const_stirring_switch_2, '0'])
                proc.communicate()
                time.sleep(3)
                proc = Popen(['/home/pi/raspberry-remote/send', const_stirring_switch_1, const_stirring_switch_2, '0'])
                proc.communicate()                    
            thread_stirring = threading.Thread(target=start_stirring)
            thread_stirring.start()

        def update_stirring_switch(*args):
            global flag_stirring
            if stirring.get() == True:
                if flag_stirring == False:
                    flag_stirring = True
                    run_stirring() # Only activate stirring if flag was false before
            else:
                flag_stirring = False
        try:
            stirring.trace_add('write', update_stirring_switch)
        except: #in case python version is lower than 3.6
            stirring.trace_variable('w', update_stirring_switch)

        global effluent_volume
        effluent_volume = tk.DoubleVar()
        with open('/home/pi/Desktop/config/Effluent.txt','r') as f:
            effluent_volume.set(float(f.read().strip()))                    
        ttk.Label(mainframe, text='Effluent V. (mL)').grid(column=column_start, row=row_start+10)
        effluent_volume_entry = ttk.Entry(mainframe, width=7, textvariable=str(effluent_volume), justify='center')
        effluent_volume_entry.grid(column=column_start+1, row=row_start+10)
        global const_max_effluent_volume
        ttk.Label(mainframe, text='Max=' + str(const_max_effluent_volume)).grid(column=column_start+2, row=row_start+10)
        
        def update_effluent_volume(*args):
            with open('/home/pi/Desktop/config/Effluent.txt','w') as f:
                try: #avoid warning messange saying '' was obtained instead of float
                    f.write(str(effluent_volume.get()))
                except:
                    pass
        try:
            effluent_volume.trace_add('write', update_effluent_volume)
        except: #in case python version is lower than 3.6
            effluent_volume.trace_variable('w', update_effluent_volume)

    def check_status(self, parent, *args, **kwargs):
        result = False
        if running_reactor == True:
            result = mbox.askyesno('Warning', 'Reactor is still running. Do you want to continue?')
        #perform action only when reactor not running or the reply for above is Yes
        if running_reactor == False or result == True:
            syringe_position()
            valve_position()
            print('PSD4 status updated.')
        return

class Status(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 11
        row_start = 0
        ttk.Label(mainframe, text='System', font='-weight bold').grid(column=column_start, row=row_start)

        ttk.Label(mainframe, text='Up time').grid(column=column_start, row=row_start+1)
        global time_script_start
        time_script_start = datetime.datetime.now()
        global script_running_time
        script_running_time = tk.StringVar()
        script_running_time.set('0:00:00')
        ttk.Label(mainframe, textvariable=script_running_time).grid(column=column_start, row=row_start+2, columnspan=2)
        def start_update_running_time():
            time.sleep(60) #give time to mainloop to start, otherwise error 'main thread is not in main loop'
            while True:
                time_diff = datetime.datetime.now()-time_script_start
                script_running_time.set(str(datetime.timedelta(seconds=round(time_diff.total_seconds()))))
                time.sleep(60)
        thread_running_time = threading.Thread(target=start_update_running_time)
        thread_running_time.start()

        global value_memory_python
        value_memory_python = tk.StringVar()
        ttk.Label(mainframe, text='Mem (MB)').grid(column=column_start, row=row_start+3)
        ttk.Label(mainframe, textvariable=value_memory_python).grid(column=column_start+1, row=row_start+3)
        global value_available_memory
        value_available_memory = tk.StringVar()
        ttk.Label(mainframe, text='Avail. (MB)').grid(column=column_start, row=row_start+4)
        ttk.Label(mainframe, textvariable=value_available_memory).grid(column=column_start+1, row=row_start+4)
        global value_cpu_temperature
        value_cpu_temperature = tk.StringVar()
        ttk.Label(mainframe, text='CPU T.').grid(column=column_start, row=row_start+5)
        ttk.Label(mainframe, textvariable=value_cpu_temperature).grid(column=column_start+1, row=row_start+5)

        global value_available_disk
        value_available_disk = tk.StringVar()
        ttk.Label(mainframe, text='Disk (GB)').grid(column=column_start, row=row_start+6)
        ttk.Label(mainframe, textvariable=value_available_disk).grid(column=column_start+1, row=row_start+6)        
        def start_update_system_status():
            time.sleep(60) #give time to mainloop to start, otherwise error 'main thread is not in main loop'
            while True:
                py = psutil.Process(os.getpid())
                memoryUse = round(py.memory_info()[0]/2**20)  # memory used by python in MB
                value_memory_python.set(str(memoryUse))
                available_memory = round(psutil.virtual_memory().available/2**20) #available memory in MB
                value_available_memory.set(available_memory)
                proc = Popen(['cat', '/sys/class/thermal/thermal_zone0/temp'], stdout=PIPE)
                out, err = proc.communicate()
                cpu_temp = round(int(out.decode())/1000, 1)
                value_cpu_temperature.set(cpu_temp)
                available_disk = round(psutil.disk_usage('/')[2]/2**30,1) #available disk in GB
                value_available_disk.set(available_disk)
                time.sleep(3)
        thread_system_info = threading.Thread(target=start_update_system_status)
        thread_system_info.start()
        
class Sampling(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 3
        row_start = 0
        ttk.Label(mainframe, text='Sampling', font='-weight bold').grid(column=column_start, row=row_start)

        ttk.Label(mainframe, text='Flush').grid(column=column_start, row=row_start+1)
        global input_flush
        input_flush = tk.IntVar()
        input_flush.set(1000)
        ttk.Entry(mainframe, width=7, textvariable=input_flush, justify='center').grid(column=column_start+1, row=row_start+1)
        
        ttk.Label(mainframe, text='Wait').grid(column=column_start, row=row_start+2)
        global input_wait
        input_wait = tk.IntVar()
        input_wait.set(10)
        ttk.Entry(mainframe, width=7, textvariable=input_wait, justify='center').grid(column=column_start+1, row=row_start+2)
        
        ttk.Label(mainframe, text='Sample volume').grid(column=column_start, row=row_start+3)
        global input_sample
        input_sample = tk.IntVar()
        input_sample.set(1500)
        ttk.Entry(mainframe, width=7, textvariable=input_sample, justify='center').grid(column=column_start+1, row=row_start+3)
        
        global button_sampling
        button_sampling = ttk.Button(mainframe, text='Sampling', width=10, command=lambda: self.start_sampling(self))
        button_sampling.grid(column=column_start+2, row=row_start+1, rowspan=3)

    def start_sampling(*args):
        def sampling():
            confirm_sampling = 'no'
            if running_reactor == True:
                confirm_sampling = mbox.askquestion('Warning', 'Reactor is still running. Make sure there is enough time for \
                        sampling before the next cycle. Yes to continue, No to abort', icon = 'warning')
                if confirm_sampling == 'yes':
                    confirm_sampling = mbox.askquestion('Attention', 'Remove the stopping cap of sampling port (3). Yes to continue, No to abort',icon = 'warning')
            else:
                confirm_sampling = mbox.askquestion('Attention', 'Remove the stopping cap of sampling port (3). Yes to continue, No to abort',icon = 'warning')
            if confirm_sampling == 'yes':
                temporary_running_flag = False
                if running_reactor == False:
                    temporary_running_flag = True #mark so that in the end running_reactor is set back to False
                running_reactor == True
                flush=int(input_flush.get())
                wait=int(input_wait.get())
                sample=int(input_sample.get())
                time_now=datetime.datetime.now()
                #flush
                pipet(const_port_reactor_out,const_port_sampling,flush,500, True) #check syringe size only for the first command
                time.sleep(wait)
                #sampling
                pipet(const_port_reactor_out,const_port_sampling,sample,500, False)
                #replenish volume
                pipet(const_port_nitrite,const_port_reactor_in, int((sample + flush)/2), 500, False)
                pipet(const_port_ammonium,const_port_reactor_in, int((sample + flush)/2), 500, False)
                #flush port_reactor_in line
                pipet(const_port_reactor_out,const_port_reactor_in, 1000, 500, False)
                #flush port_sampling line
                mbox.showinfo('Attention', 'Place sampling port (3) to DI water. Click OK to flush')
                pipet(const_port_sampling,const_port_sampling, 5000, 500, False)
                mbox.showinfo('Attention', 'Put stopping cap back to sampling port (3). Sampling is done')
                if temporary_running_flag == True:
                    running_reactor == False # see above. Set back to False if originally it was False.
                with open('/home/pi/Desktop/Reactor_log_' + time_now.strftime('%Y-%m-%d') + '.txt','a+') as f:           
                        f.write(time_now.strftime('%Y/%m/%d %H:%M:%S\t') + 'Sampling for ' + str(sample) + ' ul.\n')
                """update effluent volume"""
                effluent = float(effluent_volume.get()) + (flush+sample)/1000
                effluent_volume.set(str(effluent))
        thread_sampling = threading.Thread(target=sampling)
        thread_sampling.start()
        
class TemperatureControl(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 3
        row_start = 18
        ttk.Label(mainframe, text='Temperature').grid(column=column_start, row=row_start)
        global heating_control
        heating_control = tk.BooleanVar()
        heating_control.set(False)
        ttk.Checkbutton(mainframe, text='Controlled', variable=heating_control, command=self.start_temperature_control).grid(column=column_start+1, row=row_start)
        global heating_status
        heating_status = tk.StringVar()
        ttk.Label(mainframe, textvariable=heating_status).grid(column=column_start+2, row=row_start)
        ttk.Label(mainframe, text='Set temperature').grid(column=column_start, row=row_start+1)
        global input_temperature
        input_temperature = tk.DoubleVar()
        global const_set_point_temperature
        input_temperature.set(const_set_point_temperature)
        ttk.Entry(mainframe, width=7, textvariable=input_temperature, justify='center').grid(column=column_start+1, row=row_start+1)
        global input_temperature_tolerance
        input_temperature_tolerance = tk.DoubleVar()
        global const_meter_temperature
        if const_meter_temperature == 'DS18B20':
            input_temperature_tolerance.set(0.06)
        elif const_meter_temperature == 'Multiseven':
            input_temperature_tolerance.set(0.1)
        ttk.Entry(mainframe, width=7, textvariable=input_temperature_tolerance, justify='center').grid(column=column_start+2, row=row_start+1)
        global const_heater_switch_1
        global const_heater_switch_2
        proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '0'])
        proc.communicate()
        heating_status.set('Heating off')

    def start_temperature_control(self):
        def control_temperature():
            global input_temperature
            global input_temperature_tolerance
            global value_current_temperature
            global time_current_temperature
            global heating_control
            global const_heater_switch_1
            global const_heater_switch_2
            flag = False
            while heating_control.get():
                try:
                    temperature = input_temperature.get()
                    tolerance = input_temperature_tolerance.get()
                    current_temperature_time = time_current_temperature.get()
                    time_now = datetime.datetime.now()
                    difference = time_now - datetime.datetime.strptime(current_temperature_time, '%y/%m/%d %H:%M:%S')
                    current_temperature = float(value_current_temperature.get())
                    if difference.seconds > 60:
                        proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '0'])
                        proc.communicate()
                        heating_status.set('Heating off')
                        flag = False
                    elif current_temperature <= temperature - tolerance:
                        proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '1'])
                        proc.communicate()
                        heating_status.set('Heating on')
                        flag = True
                    elif current_temperature >= temperature + tolerance:
                        proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '0'])
                        proc.communicate()
                        heating_status.set('Heating off')
                        flag = False
                    elif flag:
                        proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '1'])
                        proc.communicate()
                        heating_status.set('Heating on')
                    else:
                        proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '0'])
                        proc.communicate()
                        heating_status.set('Heating off')
                except:
                    proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '0'])
                    proc.communicate()
                    heating_status.set('Heating off')
                    flag = False
                time.sleep(3)
            proc = Popen(['/home/pi/raspberry-remote/send', const_heater_switch_1, const_heater_switch_2, '0'])
            proc.communicate()
            heating_status.set('Heating off')
        thread_heating = threading.Thread(target=control_temperature)
        thread_heating.start()

class PressureValve(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 3
        row_start = 15
        ttk.Label(mainframe, text='Pressure valve').grid(column=column_start, row=row_start)
        ttk.Label(mainframe, text='Pressure release').grid(column=column_start, row=row_start+1)
        global valve_status
        valve_status = tk.StringVar()
        ttk.Label(mainframe, textvariable=valve_status).grid(column=column_start+2, row=row_start+1)
        ttk.Label(mainframe, text='Set threshold').grid(column=column_start, row=row_start+2)
        global const_valve_switch_1
        global const_valve_switch_2
        proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
        proc.communicate()
        valve_status.set('closed')
        global flag_releasing_gas
        flag_releasing_gas = tk.BooleanVar()
        flag_releasing_gas.set(False) #when initialized, gas release is not in process (i.e., valve is closed)
        global switch_manual_gas_valve
        switch_manual_gas_valve = tk.BooleanVar()
        switch_manual_gas_valve.set(False) # by default, do not control valve manually.
        ttk.Checkbutton(mainframe, text='Manual valve', variable=switch_manual_gas_valve).grid(column=column_start+1, row=row_start)      
        global switch_valve
        switch_valve = tk.BooleanVar()
        switch_valve.set(False)
        ttk.Checkbutton(mainframe, text='on/off', variable=switch_valve, command=self.turn_valve).grid(column=column_start+2, row=row_start)       
        global pressure_control
        pressure_control = tk.BooleanVar()
        pressure_control.set(False)
        ttk.Checkbutton(mainframe, text='Controlled', variable=pressure_control).grid(column=column_start+1, row=row_start+1)
        global input_pressure_threshold
        input_pressure_threshold = tk.DoubleVar()
        global const_pressure_threshold
        input_pressure_threshold.set(const_pressure_threshold)
        ttk.Entry(mainframe, width=7, textvariable=input_pressure_threshold, justify='center').grid(column=column_start+1, row=row_start+2)
        global input_release_time
        input_release_time = tk.DoubleVar()
        global const_pressure_release_in_seconds
        input_release_time.set(const_pressure_release_in_seconds)
        ttk.Entry(mainframe, width=7, textvariable=input_release_time, justify='center').grid(column=column_start+2, row=row_start+2)
        proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
        proc.communicate()

    def turn_valve(self):
        global running_stop_pressure
        global handle
        global const_valve_switch_1
        global const_valve_switch_2
        if switch_manual_gas_valve.get():
            if switch_valve.get():
                proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '1'])
                proc.communicate()
                valve_status.set('open ' + str(value_reactorV_imbalance.get()))
            else:
                proc = Popen(['/home/pi/raspberry-remote/send', const_valve_switch_1, const_valve_switch_2, '0'])
                proc.communicate()
                valve_status.set('closed ' + str(value_reactorV_imbalance.get()))
            return
        else:
            mbox.showinfo('Error', 'You need to check Manual valve to take control.')
            switch_valve.set(False)
            return

class Plot(tk.Frame):
    """The current plot function solved the previous problem of memory leak, but due to the new functions used in matplotlib,
        it only works on Raspberry Pi 4B but not 3B"""
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 6
        row_start = 14
        ttk.Label(mainframe, text='Plot', font='-weight bold').grid(column=column_start, row=row_start)
        # list containing all parameters that can be choosen for plotting
        global plotting
        plotting = False

        # booleans for plotting
        self.plot_pressurev = tk.BooleanVar()
        self.plot_pressurev.set(False)
        
        self.plot_pressurem = tk.BooleanVar()
        self.plot_pressurem.set(False)

        self.plot_DO2_phase = tk.BooleanVar()
        self.plot_DO2_phase.set(False)

        self.plot_DO2_ppb = tk.BooleanVar()
        self.plot_DO2_ppb.set(False)

        self.plot_temperature = tk.BooleanVar()
        self.plot_temperature.set(False)

        self.plot_pH = tk.BooleanVar()
        self.plot_pH.set(False)

        # x-value (time) for plotting
        self.time_pressure = []
        self.time_DO2 = []
        self.time_temperature_pH = []

        # y-values for plotting
        self.value_pressurev = []
        self.value_pressurem = []
        self.value_DO2_phase = []
        self.value_DO2_ppb = []
        self.value_temperature = []
        self.value_pH = []

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        # dict storing ax and all twinx needed for plotting six different sort of data
        self.axes = [self.ax, self.ax.twinx(), self.ax.twinx(), self.ax.twinx(), self.ax.twinx(), self.ax.twinx()]
        
        # list comprehension for creating a list storing all lines objects
        self.lines = [value.plot([Timestamp(0)], [0.0], 'r-')[0] for value in self.axes]

        # set position for y-axes on the right side
        for i, ax in enumerate(self.axes):
            distance = 0.1
            position = (i-1) * distance + 1
            if i > 1:
                #print(f"Position of {i}. axis is: {position}")
                # https://matplotlib.org/3.4.3/gallery/ticks_and_spines/multiple_yaxis_with_spines.html
                ax.spines.right.set_position(("axes", position))

        # dict storing all settings for design
        self.plotting_design = {
            "pressureV": {
                "checkbox": lambda : self.plot_pressurev,
                "x_ax": lambda : self.time_pressure,
                "y_ax": lambda : self.value_pressurev,
                "color": "tab:red",
                "ylabel": "pressure reactor (mV)",
                "line_style": "", # TODO this can be implemented later on
                "marker": "" # TODO this can be implemented later on
            },
            "pressureM": {
                "checkbox": lambda : self.plot_pressurem,
                "x_ax": lambda : self.time_pressure,
                "y_ax": lambda : self.value_pressurem,
                "color": "tab:orange",
                "ylabel": "pressure medium (mV)",
                "line_style": "", # TODO this can be implemented later on
                "marker": "" # TODO this can be implemented later on
            },
            "DO2 phase": {
                "checkbox": lambda : self.plot_DO2_phase,
                "x_ax": lambda : self.time_DO2,
                "y_ax": lambda : self.value_DO2_phase,
                "color": "tab:blue",
                "ylabel": "DO (phase)",
                "line_style": "", # TODO this can be implemented later on
                "marker": "" # TODO this can be implemented later on
            },
            "DO2 ppb": {
                "checkbox": lambda : self.plot_DO2_ppb,
                "x_ax": lambda : self.time_DO2,
                "y_ax": lambda : self.value_DO2_ppb,
                "color": "tab:green",
                "ylabel": "DO2 ppb",
                "line_style": "", # TODO this can be implemented later on
                "marker": "" # TODO this can be implemented later on
            },
            "temperature": {
                "checkbox": lambda : self.plot_temperature,
                "x_ax": lambda : self.time_temperature_pH,
                "y_ax": lambda : self.value_temperature,
                "color": "tab:purple",
                "ylabel": "temperature",
                "line_style": "", # TODO this can be implemented later on
                "marker": "" # TODO this can be implemented later on
            },
            "pH": {
                "checkbox": lambda : self.plot_pH,
                "x_ax": lambda : self.time_temperature_pH,
                "y_ax": lambda : self.value_pH,
                "color": "tab:gray",
                "ylabel": "pH",
                "line_style": "", # TODO this can be implemented later on
                "marker": "" # TODO this can be implemented later on
            }
        }

        self.canvas = tk.Canvas(root, width=1500, height=200, background='gray75')
        self.canvas.grid(column=0, row=1, sticky=(tk.E, tk.W, tk.N, tk.S))
        self.curve1 = FigureCanvasTkAgg(self.fig, self.canvas)
        
        global button_start_plot_continuous
        button_start_plot_continuous = ttk.Button(mainframe, width=7, text='Continuous', 
                            command=lambda: self.start_plot(mode='continuous'))
        button_start_plot_continuous.grid(column=column_start+1, row=row_start)

        global button_stop_plot_continuous
        button_stop_plot_continuous = ttk.Button(mainframe, width=7, text='Stop Plot', command=lambda: self.stop_plot())
        button_stop_plot_continuous.grid(column=column_start+1, row=row_start)
        button_stop_plot_continuous.grid_remove()

        global time_current_plot
        time_current_plot = tk.StringVar()
        ttk.Label(mainframe, textvariable=time_current_plot).grid(column=column_start+2, row=row_start)
        
        global button_start_plot_once
        button_start_plot_once = ttk.Button(mainframe, width=7, text='One time', command=lambda: self.start_plot(mode='once'))
        button_start_plot_once.grid(column=column_start+3, row=row_start)

        ttk.Label(mainframe, text='Plot recent (<24h)').grid(column=column_start, row=row_start+1)
        global input_plot_recent
        input_plot_recent = tk.DoubleVar()
        input_plot_recent.set(3)
        ttk.Entry(mainframe, width=7, textvariable=input_plot_recent, justify='center').grid(column=column_start+1, row=row_start+1)

        ttk.Label(mainframe, text='Update (s)').grid(column=column_start+2, row=row_start+1)
        global input_plot_interval
        input_plot_interval = tk.IntVar()
        input_plot_interval.set(3600)
        ttk.Entry(mainframe, width=7, textvariable=input_plot_interval, justify='center').grid(column=column_start+3, row=row_start+1)

        global mode_data_range # "From/To" for a time period between two time points, or "Recent" for the last x hours
        mode_data_range = tk.StringVar()
        # TODO changed 'Recent' to 'From/To' for testing
        mode_data_range.set('From/To')
        tk.Radiobutton(mainframe, text="Recent", variable=mode_data_range, value="Recent").grid(column=column_start, row=row_start+2)
        tk.Radiobutton(mainframe, text="From/To", variable=mode_data_range, value="From/To").grid(column=column_start+1, row=row_start+2)

        global input_plot_start #input start time of plot in %y form
        input_plot_start = tk.StringVar()
        input_plot_start.set((datetime.datetime.now()-datetime.timedelta(hours=12)).strftime('%y/%m/%d %H:%M:%S'))
        ttk.Entry(mainframe, width=15, textvariable=input_plot_start, justify='center').grid(column=column_start+2, row=row_start+2)
        global input_plot_end #input end time of plot in %y form
        input_plot_end = tk.StringVar()
        input_plot_end.set(datetime.datetime.now().strftime('%y/%m/%d %H:%M:%S'))
        ttk.Entry(mainframe, width=15, textvariable=input_plot_end, justify='center').grid(column=column_start+3, row=row_start+2)


        ttk.Checkbutton(mainframe, text='PressureV', variable=self.plot_pressurev).grid(column=column_start, row=row_start+3)
        ttk.Checkbutton(mainframe, text='PressureM', variable=self.plot_pressurem).grid(column=column_start+1, row=row_start+3)
        ttk.Checkbutton(mainframe, text='DO2 phase', variable=self.plot_DO2_phase).grid(column=column_start+2, row=row_start+3)
        ttk.Checkbutton(mainframe, text='DO2 ppb', variable=self.plot_DO2_ppb).grid(column=column_start+3, row=row_start+3)
        ttk.Checkbutton(mainframe, text='Temperature', variable=self.plot_temperature).grid(column=column_start, row=row_start+4)
        ttk.Checkbutton(mainframe, text='pH', variable=self.plot_pH).grid(column=column_start+1, row=row_start+4)

    def setup_axes(self, key, index):
        # initial set up of axes after starting plot
        # access color
        color = self.plotting_design[key]["color"]
        ax_to_plot = self.axes[index]
        ax_to_plot.set_ylabel(self.plotting_design[key]["ylabel"], color=color)
        ax_to_plot.set_xlabel("time")
        self.lines[index].set_color(color)
        ax_to_plot.tick_params(axis='y', labelcolor=color) # TODO check if this is in correct position

    def reset_plot(self):
        # set default values for axes
        for ax in self.axes:
            # hide y-label and ticks
            ax.get_yaxis().set_visible(False) # TODO this needs to be set to true when data is refereshed
            # hide spine
            right_spine = ax.spines["right"]
            right_spine.set_visible(False)
            # TODO check if labels also need to be hidden
        for line in self.lines:
            line.set_data([], [])

    def refresh_data(self, key, index, x_time, y_data_to_plot):
        # for refreshing data in continious plotting
        ax_to_plot = self.axes[index]
        ax_to_plot.set_xlim(min(x_time), max(x_time))
        ax_to_plot.set_ylim(min(y_data_to_plot), max(y_data_to_plot)) 
        # show y-label and ticks
        ax_to_plot.get_yaxis().set_visible(True)
        # show spine
        right_spine = ax_to_plot.spines["right"]
        right_spine.set_visible(True)
        self.lines[index].set_data(x_time, y_data_to_plot) 

    def draw_plot(self, mode):
        global plotting
        """"Below obtains start and end time of the plot"""
        time_start = datetime.datetime.now()
        #string_time_start = time_start.strftime('%Y/%m/%d %H:%M:%S')
        string_time_start_short = time_start.strftime('%y/%m/%d %H:%M:%S')
        #date_time_start = time_start.strftime('%Y-%m-%d')
        if mode == 'continuous': # always 'Recent' data range when 'continuous' is activated
            button_start_plot_continuous.grid_remove()
            button_stop_plot_continuous.grid()
            try:
                time_plot_start = time_start - datetime.timedelta(hours=input_plot_recent.get())
            except:
                mbox.showinfo('Error', 'Input recent time range invalid. Check format.')
                plotting = False
                return 1
            time_plot_end = time_start
        else: # plot 'once', may draw in 'Recent' range or 'From/To' range
            if mode_data_range.get() == 'From/To':
                try:
                    time_plot_start = datetime.datetime.strptime(input_plot_start.get(), '%y/%m/%d %H:%M:%S')
                    time_plot_end = datetime.datetime.strptime(input_plot_end.get(), '%y/%m/%d %H:%M:%S')
                except:
                    mbox.showinfo('Error', 'Input time format invalid. Valid format: 20/12/24 23:15:50')
                    plotting = False
                    return 1
            elif mode_data_range.get() == 'Recent':
                try:
                    time_plot_start = time_start - datetime.timedelta(hours=input_plot_recent.get())
                except:
                    mbox.showinfo('Error', 'Input recent time range invalid. Check format.')
                    plotting = False
                    return 1
                time_plot_end = time_start
        if time_plot_end - time_plot_start > datetime.timedelta(days=1):
                mbox.showinfo('Error', 'Plotting range of > 24 h is currently not supported.')
                plotting = False
                return 1
            
        while plotting == True:
            time_start = datetime.datetime.now()
            #string_time_start = time_start.strftime('%Y/%m/%d %H:%M:%S')
            string_time_start_short = time_start.strftime('%y/%m/%d %H:%M:%S')
            #date_time_start = time_start.strftime('%Y-%m-%d')
            """get 'interval', only useful for continuous mode"""
            interval = 3600 #intialize the value
            try:
                interval = input_plot_interval.get()
            except:
                print('Warning: Invalid interval (s) value. A default value of 3600 is used.')
                interval = 3600
            time_end = time_start + datetime.timedelta(seconds=interval)

            """get data""" # TODO instead of using threads maybe use asyncio. Async opening of files: https://www.twilio.com/blog/working-with-files-asynchronously-in-python-using-aiofiles-and-asyncio
            if self.plot_pressurev.get() or self.plot_pressurem.get():
                try:      
                    """get the files for pressure"""
                    if time_plot_start.date() == datetime.datetime.now().date():
                        file_pressure1 = r'/home/pi/Desktop/T7Pro_log_' + time_plot_start.strftime('%Y-%m-%d') + '.txt'
                    else:
                        file_pressure1 = r'/home/pi/Desktop/log/' + time_plot_start.strftime('%Y-%m-%d') + '/T7Pro_log_' + time_plot_start.strftime('%Y-%m-%d') + '.txt'
                    if time_plot_end.date() == datetime.datetime.now().date():
                        file_pressure2 = r'/home/pi/Desktop/T7Pro_log_' + time_plot_end.strftime('%Y-%m-%d') + '.txt'
                    else:
                        file_pressure2 = r'/home/pi/Desktop/log/' + time_plot_end.strftime('%Y-%m-%d') + '/T7Pro_log_' + time_plot_end.strftime('%Y-%m-%d') + '.txt'
                    """read file with pandas, specify all the columns present in the file"""
                    df_pressure1 = read_csv(file_pressure1, sep="\t|=", names=['time','Redox_label', 'Redox', 'PR_label','Pressure_Reactor', 'PM_label','Pressure_Medium'], \
                                    parse_dates=['time'], infer_datetime_format=True, engine='python')
                    if file_pressure1 != file_pressure2:               
                        df_pressure2 = read_csv(file_pressure2, sep="\t|=", names=['time','Redox_label', 'Redox', 'PR_label','Pressure_Reactor', 'PM_label','Pressure_Medium'], \
                                        parse_dates=['time'], infer_datetime_format=True, engine='python')
                        df_pressure = concat([df_pressure1, df_pressure2])
                    else:
                        df_pressure = df_pressure1
                    mask = (df_pressure['time'] > time_plot_start.strftime('%Y/%m/%d %H:%M:%S')) & (df_pressure['time'] < time_plot_end.strftime('%Y/%m/%d %H:%M:%S'))
                    time_pressure = df_pressure.loc[mask]["time"]
                    self.time_pressure = time_pressure.tolist()
                    if self.plot_pressurev.get():
                        value_pressurev = df_pressure.loc[mask]["Pressure_Reactor"]
                        self.value_pressurev = value_pressurev.tolist()
                    if self.plot_pressurem.get():
                        value_pressurem = df_pressure.loc[mask]["Pressure_Medium"]
                        self.value_pressurem = value_pressurem.tolist()
                except:
                    mbox.showinfo('Error', 'Error with fetching pressure data.')
                    plotting = False
                    return 1
                
            if self.plot_DO2_phase.get() or self.plot_DO2_ppb.get():
                try:
                    """get the files for DO2"""
                    if time_plot_start.date() == datetime.datetime.now().date():
                        file_DO2_1 = r'/home/pi/Desktop/Oxygen2_log_' + time_plot_start.strftime('%Y-%m-%d') + '.txt'
                    else:
                        file_DO2_1 = r'/home/pi/Desktop/log/' + time_plot_start.strftime('%Y-%m-%d') + '/Oxygen2_log_' + time_plot_start.strftime('%Y-%m-%d') + '.txt'
                    if time_plot_end.date() == datetime.datetime.now().date():
                        file_DO2_2 = r'/home/pi/Desktop/Oxygen2_log_' + time_plot_end.strftime('%Y-%m-%d') + '.txt'
                    else:
                        file_DO2_2 = r'/home/pi/Desktop/log/' + time_plot_end.strftime('%Y-%m-%d') + '/Oxygen2_log_' + time_plot_end.strftime('%Y-%m-%d') + '.txt'
                    """read file with pandas, specify all the columns present in the file"""
                    df_DO2_1 = read_csv(file_DO2_1, sep="\t", names=['time','amplitude_label', 'amplitude', 'phase_label','phase', 'ppb_label','ppb', 'comment'], \
                                    parse_dates=['time'], infer_datetime_format=True, engine='python')
                    if file_DO2_1 != file_DO2_2:
                        df_DO2_2 = read_csv(file_DO2_2, sep="\t", names=['time','amplitude_label', 'amplitude', 'phase_label','phase', 'ppb_label','ppb', 'comment'], \
                                        parse_dates=['time'], infer_datetime_format=True, engine='python')
                        df_DO2 = concat([df_DO2_1, df_DO2_2])
                    else:
                        df_DO2 = df_DO2_1

                    mask = (df_DO2['time'] > time_plot_start.strftime('%Y/%m/%d %H:%M:%S')) & (df_DO2['time'] < time_plot_end.strftime('%Y/%m/%d %H:%M:%S'))
                    time_DO2 = df_DO2.loc[mask]["time"]
                    self.time_DO2 = time_DO2.tolist()
                    if self.plot_DO2_phase.get():
                        value_DO2_phase = df_DO2.loc[mask]["phase"]
                        self.value_DO2_phase = value_DO2_phase.tolist()
                    if self.plot_DO2_ppb.get():
                        value_DO2_ppb = df_DO2.loc[mask]["ppb"]
                        self.value_DO2_ppb = value_DO2_ppb.tolist()
                except:
                    mbox.showinfo('Error', 'Error with fetching DO data.')
                    plotting = False
                    return 1

            if self.plot_temperature.get() or self.plot_pH.get():
                try:        
                    """get the files for temperature/pH"""
                    if time_plot_start.date() == datetime.datetime.now().date():
                        file_temperature1 = r'/home/pi/Desktop/Temperature_log_' + time_plot_start.strftime('%Y-%m-%d') + '.txt'
                    else:
                        file_temperature1 = r'/home/pi/Desktop/log/' + time_plot_start.strftime('%Y-%m-%d') + '/Temperature_log_' + time_plot_start.strftime('%Y-%m-%d') + '.txt'
                    if time_plot_end.date() == datetime.datetime.now().date():
                        file_temperature2 = r'/home/pi/Desktop/Temperature_log_' + time_plot_end.strftime('%Y-%m-%d') + '.txt'
                    else:
                        file_temperature2 = r'/home/pi/Desktop/log/' + time_plot_end.strftime('%Y-%m-%d') + '/Temperature_log_' + time_plot_end.strftime('%Y-%m-%d') + '.txt'
                    """read file with pandas, specify all the columns present in the file"""
                    df_temperature1 = read_csv(file_temperature1, sep="\t|=", names=['time','pH_label', 'pH', 'temperature_label','temperature'], \
                                    parse_dates=['time'], infer_datetime_format=True, engine='python')
                    if file_temperature1 != file_temperature2:               
                        df_temperature2 = read_csv(file_temperature2, sep="\t|=", names=['time','pH_label', 'pH', 'temperature_label','temperature'], \
                                        parse_dates=['time'], infer_datetime_format=True, engine='python')
                        df_temperature = concat([df_temperature1, df_temperature2])
                    else:
                        df_temperature = df_temperature1
                    mask = (df_temperature['time'] > time_plot_start.strftime('%Y/%m/%d %H:%M:%S')) & (df_temperature['time'] < time_plot_end.strftime('%Y/%m/%d %H:%M:%S'))
                    time_temperature_pH = df_temperature.loc[mask]["time"]
                    self.time_temperature_pH = time_temperature_pH.tolist()

                    if self.plot_temperature.get():
                        value_temperature = df_temperature.loc[mask]["temperature"]
                        self.value_temperature = value_temperature.tolist()
                    if self.plot_pH.get():
                        value_pH = df_temperature.loc[mask]["pH"]
                        self.value_pH = value_pH.tolist()
                except:
                    mbox.showinfo('Error', 'Error with fetching temperature/pH data.')
                    plotting = False
                    return 1
                
            if self.plot_pressurev.get() + self.plot_pressurem.get() + self.plot_DO2_phase.get() + self.plot_DO2_ppb.get() \
                + self.plot_temperature.get() + self.plot_pH.get() == False:
                mbox.showinfo('Error', 'No parameters were selected.')
                plotting = False
                return 1
                
            """plot"""
            # check which checkboxes are chosen
            index = 0
            # reset axes
            self.reset_plot()
            for key in self.plotting_design: 
                if self.plotting_design[key]["checkbox"]().get():
                    try:
                        # do design setup for checked data
                        self.setup_axes(key=key, index=index)
                        # refresh data for checked data
                        self.refresh_data(key=key, index=index, x_time=self.plotting_design[key]["x_ax"](), y_data_to_plot=self.plotting_design[key]["y_ax"]())
                        index += 1 # TODO do I need to remember indices for continious plotting
                    except:
                        mbox.showinfo('Error', 'Error with refrehsing data.')
                        plotting = False
                        return 1
            
            if mode == 'continuous':
                save_path = r'/home/pi/Desktop/plot/continuous/Plot_' + time_start.strftime('%Y-%m-%d-%H-%M-%S') + '.png'
            elif mode == 'once':
                save_path = r'/home/pi/Desktop/plot/once/Plot_' + time_start.strftime('%Y-%m-%d-%H-%M-%S') + '.png'

            self.fig.savefig(save_path,bbox_inches='tight')
            # canvas.delete('all') #clear the canvas
            
            self.curve1.draw()
            a=self.curve1.get_tk_widget()
            a.grid(column=0, row=0, sticky=(tk.E, tk.W, tk.N, tk.S))
            self.canvas.create_window(0, 0, width=1200, height=200, anchor='nw', window=a)
            time_current_plot.set(string_time_start_short)
            
            if mode == 'once':
                plotting = False #this will cause the below while loop to exit (becaue interval = 0)
                # plt.close('all')
            flag = 1 #force into while loop at least once
            while flag ==1 or datetime.datetime.now() < time_end:
                flag = 0
                time.sleep(3)
                if plotting:
                    try:
                        interval = input_plot_interval.get()
                    except:
                        interval = interval + 4 #make it longer than the 3s sleep
                else:
                    interval = 0
                time_end = time_start + datetime.timedelta(seconds=interval)
        if mode == 'continuous':
            button_start_plot_continuous.grid()
            button_stop_plot_continuous.grid_remove()

    def start_plot(self, mode="once"): #mode is 'continuous' or 'once'
        global plotting
        global plot_DO2_phase
        global plot_DO2_ppb
        global plot_temperature
        global plot_pH
        if plotting == True:
            mbox.showinfo('Error', 'Please stop continuous plotting before making a one-time plot.')
            return 1
        plotting = True
        thread_plotting = threading.Thread(target= lambda : self.draw_plot(mode))
        thread_plotting.start()

    def stop_plot(self):
        global plotting
        plotting = False

class Slope(tk.Frame):
    def __init__(self, parent, *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        self.parent = parent
        column_start = 6
        row_start = 19
        ttk.Label(mainframe, text='Slope', font='-weight bold').grid(column=column_start, row=row_start)
        global flag_slope_refresh
        flag_slope_refresh = tk.BooleanVar()
        flag_slope_refresh.set(False)

        global slope1_value1
        global slope2_value1
        global slope3_value1
        global slope4_value1
        global slope5_value1
        global slope6_value1
        global slope7_value1
        global slope8_value1
        global slope9_value1
        global slope1_value2
        global slope2_value2
        global slope3_value2
        global slope4_value2
        global slope5_value2
        global slope6_value2
        global slope7_value2
        global slope8_value2
        global slope9_value2
        global slope1_value3
        global slope2_value3
        global slope3_value3
        global slope4_value3
        global slope5_value3
        global slope6_value3
        global slope7_value3
        global slope8_value3
        global slope9_value3
        global slope1_value4
        global slope2_value4
        global slope3_value4
        global slope4_value4
        global slope5_value4
        global slope6_value4
        global slope7_value4
        global slope8_value4
        global slope9_value4
        global slope1_value5
        global slope2_value5
        global slope3_value5
        global slope4_value5
        global slope5_value5
        global slope6_value5
        global slope7_value5
        global slope8_value5
        global slope9_value5
        global slope1_value6
        global slope2_value6
        global slope3_value6
        global slope4_value6
        global slope5_value6
        global slope6_value6
        global slope7_value6
        global slope8_value6
        global slope9_value6
        global slope1_value7
        global slope2_value7
        global slope3_value7
        global slope4_value7
        global slope5_value7
        global slope6_value7
        global slope7_value7
        global slope8_value7
        global slope9_value7
        slope1_value1 = tk.StringVar()
        slope2_value1 = tk.StringVar()
        slope3_value1 = tk.StringVar()
        slope4_value1 = tk.StringVar()
        slope5_value1 = tk.StringVar()
        slope6_value1 = tk.StringVar()
        slope7_value1 = tk.StringVar()
        slope8_value1 = tk.StringVar()
        slope9_value1 = tk.StringVar()
        slope1_value3 = tk.StringVar()
        slope2_value3 = tk.StringVar()
        slope3_value3 = tk.StringVar()
        slope4_value3 = tk.StringVar()
        slope5_value3 = tk.StringVar()
        slope6_value3 = tk.StringVar()
        slope7_value3 = tk.StringVar()
        slope8_value3 = tk.StringVar()
        slope9_value3 = tk.StringVar()
        slope1_value2 = tk.StringVar()
        slope2_value2 = tk.StringVar()
        slope3_value2 = tk.StringVar()
        slope4_value2 = tk.StringVar()
        slope5_value2 = tk.StringVar()
        slope6_value2 = tk.StringVar()
        slope7_value2 = tk.StringVar()
        slope8_value2 = tk.StringVar()
        slope9_value2 = tk.StringVar()
        slope1_value4 = tk.StringVar()
        slope2_value4 = tk.StringVar()
        slope3_value4 = tk.StringVar()
        slope4_value4 = tk.StringVar()
        slope5_value4 = tk.StringVar()
        slope6_value4 = tk.StringVar()
        slope7_value4 = tk.StringVar()
        slope8_value4 = tk.StringVar()
        slope9_value4 = tk.StringVar()
        slope1_value5 = tk.StringVar()
        slope2_value5 = tk.StringVar()
        slope3_value5 = tk.StringVar()
        slope4_value5 = tk.StringVar()
        slope5_value5 = tk.StringVar()
        slope6_value5 = tk.StringVar()
        slope7_value5 = tk.StringVar()
        slope8_value5 = tk.StringVar()
        slope9_value5 = tk.StringVar()
        slope1_value6 = tk.StringVar()
        slope2_value6 = tk.StringVar()
        slope3_value6 = tk.StringVar()
        slope4_value6 = tk.StringVar()
        slope5_value6 = tk.StringVar()
        slope6_value6 = tk.StringVar()
        slope7_value6 = tk.StringVar()
        slope8_value6 = tk.StringVar()
        slope9_value6 = tk.StringVar()
        slope1_value7 = tk.StringVar()
        slope2_value7 = tk.StringVar()
        slope3_value7 = tk.StringVar()
        slope4_value7 = tk.StringVar()
        slope5_value7 = tk.StringVar()
        slope6_value7 = tk.StringVar()
        slope7_value7 = tk.StringVar()
        slope8_value7 = tk.StringVar()
        slope9_value7 = tk.StringVar()


        global time_last_release
        global value_pressure_last_release
        global value_current_pressurev
        global time_current_pressurev

        """set label and position in batch using eval"""
        for i in range (0,9):
            for j in range (0,7):
                ttk.Label(mainframe, textvariable=eval('slope' + str(i+1) + '_value' + str(j+1))).grid(column=column_start+j, row=row_start+i+1)

        def update_slope(*args): # *args is necessary although pylance complains
            if flag_slope_refresh.get()==False and flag_releasing_gas.get()==False:
                """only update the last line when there is a new pressure reading 
                    and when gas release is not in progress"""
                f = open('/home/pi/Desktop/config/Slope.txt', 'rb')
                slope_text = tail(f,1)
                t = slope_text.split('\t')
                time_last_release = t[0][2:]
                value_pressure_last_release = t[1]
            elif flag_slope_refresh.get()==True:
                """update the first eight lines after a gas release is finished (therefore 
                    the flag_slope_refresh is true)"""
                f = open('/home/pi/Desktop/config/Slope.txt', 'rb')
                slope_text = tail(f,8)
                t = slope_text.split('\n')
                for i in range (0,8):
                    try:
                        s = t[i].split('\t')
                        time_slope_start = datetime.datetime.strptime(s[0], '%Y/%m/%d %H:%M:%S')
                        eval('slope' + str(i+1) + '_value1').set(time_slope_start.strftime("%y/%m/%d %H:%M:%S"))
                        eval('slope' + str(i+1) + '_value2').set(round(float(s[1]),1))
                        time_slope_end = datetime.datetime.strptime(s[2], '%Y/%m/%d %H:%M:%S')
                        eval('slope' + str(i+1) + '_value3').set(time_slope_end.strftime("%y/%m/%d %H:%M:%S"))
                        eval('slope' + str(i+1) + '_value4').set(round(float(s[3]),1))
                        eval('slope' + str(i+1) + '_value5').set(str(round(float(s[3])-float(s[1]),1)) + ' mV')
                        time_diff = time_slope_end - time_slope_start
                        eval('slope' + str(i+1) + '_value6').set(time_diff)
                        eval('slope' + str(i+1) + '_value7').set(str(round((float(s[3])-float(s[1]))/(float(time_diff.seconds)/86400 + time_diff.days),1)) + ' mV/d')
                    except:
                        pass
            try:
                slope9_value1.set(time_last_release)
                slope9_value2.set(round(float(value_pressure_last_release),1))
                time_current_pressure = time_current_pressurev.get()
                current_pressure = value_current_pressurev.get()
                slope9_value3.set(time_current_pressure)
                slope9_value4.set(round(float(current_pressure),1))
                time_diff = datetime.datetime.strptime(time_current_pressure, '%y/%m/%d %H:%M:%S')-datetime.datetime.strptime(time_last_release, '%y/%m/%d %H:%M:%S')
                slope9_value5.set(str(round(float(current_pressure)-float(value_pressure_last_release),1)) + ' mV')
                slope9_value6.set(time_diff)
                slope9_value7.set(str(round((float(current_pressure)-float(value_pressure_last_release))/(float(time_diff.seconds)/86400 + time_diff.days),1)) + ' mV/d')
            except:
                pass
        """Initilize all the nine lines by reading from Slope.txt"""
        f = open('/home/pi/Desktop/config/Slope.txt', 'rb')
        slope_text = tail(f,9)
        t = slope_text.split('\n')
        for i in range (0,9):
            try:
                s = t[i].split('\t')
                time_slope_start = datetime.datetime.strptime(s[0], '%Y/%m/%d %H:%M:%S')
                eval('slope' + str(i+1) + '_value1').set(time_slope_start.strftime("%y/%m/%d %H:%M:%S"))
                eval('slope' + str(i+1) + '_value2').set(round(float(s[1]),1))
                time_slope_end = datetime.datetime.strptime(s[2], '%Y/%m/%d %H:%M:%S')
                eval('slope' + str(i+1) + '_value3').set(time_slope_end.strftime("%y/%m/%d %H:%M:%S"))
                eval('slope' + str(i+1) + '_value4').set(round(float(s[3]),1))
                eval('slope' + str(i+1) + '_value5').set(str(round(float(s[3])-float(s[1]),1)) + ' mV')
                time_diff = time_slope_end - time_slope_start
                eval('slope' + str(i+1) + '_value6').set(time_diff)
                eval('slope' + str(i+1) + '_value7').set(str(round((float(s[3])-float(s[1]))/(float(time_diff.seconds)/86400 + time_diff.days),1)) + ' mV/d')
            except:
                pass

        try:
            """after new slope was logged in (after gas release), update slope"""
            flag_slope_refresh.trace_add('write', update_slope)
            """after new reading of pressure, update the last line of slope"""
            time_current_pressurev.trace_add('write', update_slope)
        except: #in case python version is lower than 3.6
            flag_slope_refresh.trace_variable('w', update_slope)
            time_current_pressurev.trace_variable('w', update_slope)


if __name__ == '__main__':
    fileDirectory = os.path.dirname(os.path.abspath(__file__)) #therefore this python code has to be placed in the same directory as other folders (log, config, screenshot, plot)
    #f = open(os.path.join(fileDirectory, 'log', '2020-12-01', 'Temperature_log_2020-12-01.txt'), 'rb')
    #print (tail(f,10))
    root = tk.Tk()    #tkinter.Tk is the class that create the window. The window host tkinter.Frame and tkinter.ttk.Frame
    MainApplication(root)
    root.mainloop()