"""built-in modules"""
import time
import datetime
import re
import os
from subprocess import Popen, PIPE # these commands does not require shell=True: mkdir/cat/rm/dropbox_uploader/'sudo teamviewer daemon restart'
import glob #used in uploading files
import threading
from math import log #used in calcualting doubling time
from tkinter import ttk
import tkinter as tk
import tkinter.messagebox as mbox
"""third party modules"""
import serial
from serial.tools import list_ports
import minimalmodbus #for oxy1rs485

#Run parity_change() if the OXY1RS485 device is configured to have parity EVEN

global const_oxy1rs485_port

"""determine OXY1RS485 port"""
ports = list_ports.comports() #get all ports
port_oxy1rs485 = '-'
for p in ports:
    if "AV0LFSCG" in p.hwid: # specific ID of the connection cable (but not the device). E.g., in the format of AV0LFSCG or FT52EEW1
        port_oxy1rs485 = p.device
const_oxy1rs485_port = port_oxy1rs485

print(const_oxy1rs485_port)

def parity_change():
    """This changes parity from EVEN (0) to NONE (2)"""
    global const_oxy1rs485_port 
    instrument = minimalmodbus.Instrument(const_oxy1rs485_port,1) #1 is Modbus ID
    instrument.serial.parity = "E"
    instrument.serial.stopbits = 1
    instrument.serial.timeout = 0.2    
    #type = instrument.read_long(4925,byteorder = 3)
    instrument.write_registers(4107, [0x02,0x00])
    instrument.serial.close() # the port needs to be closed. Otherwise "error from flowcontrol urb" occurs (dmesg | grep tty) after some time, and usb port changes
    return

def read_oxy1rs485():
    """let oxy1rs485 read and return a string that contains amplitude, phase, and ppb"""
    global const_oxy1rs485_port 
    instrument = minimalmodbus.Instrument(const_oxy1rs485_port,1) #1 is Modbus ID
    instrument.serial.timeout = 1.0
    instrument.serial.parity = "N"
    instrument.serial.timeout = 0.2
    instrument.write_registers(9000, [0x04,0x00]) #Trigger single measurement, see manual EOM-_t_O2-FOM_Modbus_RTU_Communication_Protocol
    time.sleep(1) #1s waiting is necessary for measurement to finish
    amplitude = instrument.read_float(4899,byteorder = 3)
    phase = instrument.read_float(4901,byteorder = 3)*100
    ppb = instrument.read_float(4905,byteorder = 3)
    instrument.serial.close() # the port needs to be closed. Otherwise "error from flowcontrol urb" occurs (dmesg | grep tty) after some time, and usb port changes
    return str(round(amplitude,0)) + "-" + str(round(phase,3)) + "-" + str(round(ppb,3))

def change_sensor():
    """let oxy1rs485 change sensor type: PST3, PST6, and PST9"""
    global const_oxy1rs485_port 
    instrument = minimalmodbus.Instrument(const_oxy1rs485_port,1) #1 is Modbus ID
    instrument.serial.parity = "N"
    instrument.serial.timeout = 0.2    
    instrument.write_registers(4925, [0x02,0x00]) #Change sensor type (0x02: PST3, 0x03: PST6, 0x06: PST9), see manual EOM-_t_O2-FOM_Modbus_RTU_Communication_Protocol
    instrument.serial.close() # the port needs to be closed. Otherwise "error from flowcontrol urb" occurs (dmesg | grep tty) after some time, and usb port changes
    return

